""" Represents a data upload task. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: data_upload_task.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import asyncio
import logging
import tempfile
import time
from datetime import datetime
from os.path import join
from pathlib import Path
from typing import Any, Callable

from PySide6 import QtCore
from PySide6.QtGui import QImage, QPixmap

from pasta_eln.backend import Backend
from pasta_eln.dataverse.client import DataverseClient
from pasta_eln.dataverse.config_error import ConfigError
from pasta_eln.dataverse.config_model import ConfigModel
from pasta_eln.dataverse.database_api import DatabaseAPI
from pasta_eln.dataverse.generic_task_object import GenericTaskObject
from pasta_eln.dataverse.progress_updater_thread import ProgressUpdaterThread
from pasta_eln.dataverse.upload_model import UploadModel
from pasta_eln.dataverse.upload_status_values import UploadStatusValues
from pasta_eln.dataverse.utils import get_flattened_metadata, update_status
from pasta_eln.inputOutput import exportELN


class DataUploadTask(GenericTaskObject):
  """
  Represents a task for uploading data to a Dataverse repository.

  Explanation:
      This class manages the process of uploading data to Dataverse, including creating datasets,
      uploading files, checking dataset locks, and handling cancellation of the upload task.
  """
  progress_changed = QtCore.Signal(int)
  status_changed = QtCore.Signal(str)
  upload_model_created = QtCore.Signal(str)

  def __init__(self,
               project_name: str,
               project_doc_id: str,
               progress_update_callback: Callable[[int], None],
               status_label_set_text_callback: Callable[[str], None],
               status_icon_set_pixmap_callback: Callable[[QPixmap | QImage | str], None],
               upload_cancel_clicked_signal_callback: QtCore.Signal,
               backend: Backend) -> None:
    """
    Initializes the DataUploadTask with necessary callbacks and information.

    Args:
        project_name (str): The name of the project for the upload task.
        progress_update_callback (Callable[[int], None]): Callback for updating progress.
        status_label_set_text_callback (Callable[[str], None]): Callback for setting status label text.
        status_icon_set_pixmap_callback (Callable[[QPixmap | QImage | str], None]): Callback for setting status icon.
        upload_cancel_clicked_signal_callback (QtCore.Signal): Signal for upload cancellation.
    """
    super().__init__()
    self.db_api: DatabaseAPI | None = None
    self.backend = backend
    self.dataverse_client: DataverseClient | None = None
    self.upload_model: UploadModel | None = None
    self.config_model: ConfigModel | None = None
    self.dataverse_id: str = ""
    self.dataverse_api_token: str = ""
    self.dataverse_server_url: str = ""
    self.metadata: dict[str, Any] = {}

    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.project_name = project_name
    self.project_doc_id = project_doc_id

    # Connect slots
    self.progress_changed.connect(progress_update_callback)
    self.status_changed.connect(
      lambda status: update_status(status, status_label_set_text_callback, status_icon_set_pixmap_callback))
    upload_cancel_clicked_signal_callback.connect(self.cancel.emit)  # type: ignore[attr-defined]

    # Create progress thread to update the progress
    self.progress_thread = ProgressUpdaterThread()
    self.progress_thread.progress_update = self.progress_changed
    self.progress_thread.end.connect(self.progress_thread.quit)

  def start_task(self) -> None:
    """
    Start the data upload task by generating an ELN file, creating a dataset in Dataverse,
    uploading the file, and finalizing the upload task.

    Args:
        self: Instance of the data upload task.
    """
    super().start_task()
    self.initialize()

    # Emitting signals to update initial status
    self.update_changed_status(UploadStatusValues.Uploading.name)
    # Emit the upload model id
    # so that the parent thread can retrieve the model
    if self.upload_model:
      self.upload_model_created.emit(self.upload_model.id)

    # Start the progress updater thread
    self.progress_thread.start()

    # Step 1: Generate ELN file for the project
    if self.check_if_cancelled():
      return
    with tempfile.TemporaryDirectory() as tmp_dir:
      eln_file_path = self.generate_eln_file(tmp_dir)
      if not eln_file_path:
        self.update_log(f"Failed to generate ELN file for project: {self.project_name}, hence finalizing the upload",
                        self.logger.warning)
        self.finalize_upload_task(UploadStatusValues.Error.name)
        return
      self.update_log(f"Successfully generated ELN file: {eln_file_path}", self.logger.info)

      # Step 2: Create dataset for the project in dataverse
      if self.check_if_cancelled():
        return
      persistent_id = self.create_dataset_for_pasta_project()
      if persistent_id is None:
        self.update_log(f"Failed to create dataset for project: {self.project_name}, hence finalizing the upload",
                        self.logger.warning)
        self.finalize_upload_task(UploadStatusValues.Error.name)
        return

      # Step 3: Check if the dataset is unlocked post the creation
      if self.check_if_cancelled():
        return
      if not self.check_if_dataset_is_unlocked(persistent_id):
        self.update_log(f"Failed to unlock dataset for project: {self.project_name}, hence finalizing the upload",
                        self.logger.warning)
        self.finalize_upload_task(UploadStatusValues.Error.name)
        return

      # Step 4: Upload the generated ELN file to the dataset
      if self.check_if_cancelled():
        return
      file_pid = self.upload_generated_eln_file_to_dataset(persistent_id, eln_file_path)
      if file_pid is None:
        self.update_log(
          f"Failed to upload eln file to dataset for project: {self.project_name}, hence finalizing the upload",
          self.logger.warning)
        self.finalize_upload_task(UploadStatusValues.Error.name)
        return

      # Final step: Update the log and finalize the upload task
      upload_url = f"{self.dataverse_server_url}/dataset.xhtml?persistentId={persistent_id}"
      self.update_log(
        f"Successfully uploaded ELN file, URL: {upload_url}",
        self.logger.info)
      if self.upload_model:
        self.upload_model.dataverse_url = upload_url
      self.finalize_upload_task(
        UploadStatusValues.Cancelled.name if self.cancelled else UploadStatusValues.Finished.name)

  def initialize(self) -> None:
    """
    Initialize the data upload task by setting up the database API,
    creating an upload model, and reading configuration information.
    Since this method consists of heavy operations, it's not invoked inside the constructor.

    Raises:
        ConfigError: If the config model is not found or contains invalid login information.
    """

    self.db_api = DatabaseAPI()
    # Create upload model in database
    self.upload_model = UploadModel(project_name=self.project_name,
                                    status=UploadStatusValues.Queued.name,
                                    log=f"Upload initiated for project {self.project_name} at {datetime.now().isoformat()}\n",
                                    created_date_time=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                                    project_doc_id=self.project_doc_id)
    self.upload_model = self.db_api.create_model_document(self.upload_model)  # type: ignore[assignment]
    self.logger.info("Upload model created: %s", self.upload_model)
    # Read config and get the login information along with the metadata
    self.config_model = self.db_api.get_config_model()
    if self.config_model is None:
      raise ConfigError("Config model not found/Invalid Login Information.")
    login_info = self.config_model.dataverse_login_info or {}
    self.dataverse_server_url = login_info.get("server_url", "")
    self.dataverse_api_token = login_info.get("api_token", "")
    self.dataverse_id = login_info.get("dataverse_id", "")
    self.metadata = self.config_model.metadata or {}
    self.dataverse_client = DataverseClient(self.dataverse_server_url, self.dataverse_api_token, 60)

  def finalize_upload_task(self, status: str = UploadStatusValues.Finished.name) -> None:
    """
    Finalizes the upload task by emitting signals to cancel progress, mark as finish, and update the UI status.

    Args:
        status (str): The status to be emitted.
    """
    self.progress_thread.finalize.emit()
    if self.upload_model:
      self.upload_model.finished_date_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    self.update_changed_status(status)
    self.finish.emit()

  def create_dataset_for_pasta_project(self) -> str | None:
    """
    Creates a dataset for the PASTA project by adjusting metadata and invoking the dataverse client method.

    Returns:
        str or None: The persistent identifier of the created dataset if successful, None otherwise.
    """
    if self.upload_model is None or self.dataverse_client is None:
      return None
    self.update_log(f"Step 2: Creating dataset..... {self.upload_model.project_name}", self.logger.info)
    persistent_id = None

    # Get the saved metadata from the database
    # and adjust it to the required format for invoking the dataverse client method
    metadata_adjusted = get_flattened_metadata(self.metadata)
    metadata_adjusted["title"] = self.upload_model.project_name  # Set the title to the PASTA project name
    result = asyncio.run(self.dataverse_client.create_and_publish_dataset(self.dataverse_id, metadata_adjusted))
    if (isinstance(result, dict) and
        'identifier' in result and
        'authority' in result and
        'protocol' in result):
      persistent_id = f"{result.get('protocol')}:{result.get('authority')}/{result.get('identifier')}"
      log = f'Dataset creation succeeded with PID: {persistent_id}'
      self.update_log(log, self.logger.info)
      self.logger.info(log)
    else:
      self.update_log(f'Dataset creation failed with errors: {result}', self.logger.error)
    return persistent_id

  def check_if_dataset_is_unlocked(self, persistent_id: str) -> bool:
    """
    Checks if a dataset is unlocked after a publish operation.

    Args:
        persistent_id (str): The persistent identifier of the dataset to check.

    Returns:
        bool: True if the dataset is unlocked, False otherwise.
    """
    if self.dataverse_client is None:
      return False
    self.update_log("Step 3: Checking if dataset is unlocked after the publish operation..", self.logger.info)
    unlocked = False
    for _ in range(10):
      result = asyncio.run(self.dataverse_client.get_dataset_locks(persistent_id))
      if isinstance(result, dict):
        if locks := result.get('locks'):
          for lock in locks:
            self.update_log(
              f"Dataset with PID: {persistent_id} is locked. Lock type: {lock.get('lockType')}, Message: {lock.get('message')}",
              self.logger.info)
        else:
          self.update_log(f"Dataset with PID: {persistent_id} is unlocked already!", self.logger.info)
          unlocked = True
          break
      else:
        self.update_log(f"Dataset lock check failed. {result}", self.logger.info)
      time.sleep(1)
    if not unlocked:
      self.update_log(f"Dataset with PID: {persistent_id} is still locked after 10 retries!", self.logger.error)
    return unlocked

  def upload_generated_eln_file_to_dataset(self, persistent_id: str, eln_file_path: str) -> str | None:
    """
    Uploads the generated ELN file to a dataset in Dataverse.

    Args:
        persistent_id (str): The persistent identifier of the dataset.
        eln_file_path (str): The path to the ELN file to upload.

    Returns:
        str or None: The file PID if upload is successful, None otherwise.
    """
    self.update_log(f"Step 4: Uploading file {eln_file_path} to dataset with PID: {persistent_id}.....",
                    self.logger.info)
    if self.upload_model is None or self.dataverse_client is None:
      return None
    result = asyncio.run(self.dataverse_client.upload_file(persistent_id,
                                                           eln_file_path,
                                                           f"Generated ELN file for the project: {self.upload_model.project_name}",
                                                           ["ELN"]))
    if (isinstance(result, dict)
        and 'file_upload_result' in result
        and 'dataset_publish_result' in result):
      dataset_publish_result = result["dataset_publish_result"]
      file_pid = f"{dataset_publish_result.get('protocol')}:{dataset_publish_result.get('authority')}/{dataset_publish_result.get('identifier')}"
      self.update_log(f'ELN File uploaded successfully, generated file PID: {file_pid}', self.logger.info)
    else:
      self.update_log(f'ELN File upload failed with errors: {result}', self.logger.error)
      return None
    return file_pid

  def update_log(self, log: str, logger_method: Callable[[str], None] | None) -> None:
    """
    Updates the log with the provided message and triggers the logger method if available.

    Args:
        log (str): The log message to be updated.
        logger_method (Callable[[str], None] | None): The method to handle logging.
    """
    if logger_method:
      logger_method(log)
    if self.upload_model and self.db_api:
      self.upload_model.log = log
      self.db_api.update_model_document(self.upload_model)

  def cancel_task(self) -> None:
    """
    Cancels the data upload task.

    Explanation:
        This method cancels the data upload task by calling the super-class's cancel_task method.
        It updates the upload model log and status accordingly, and emits the status_changed signal.

    """
    if self.cancelled or self.finished:
      return
    super().cancel_task()
    if self.upload_model:
      self.upload_model.log = f"Cancelled at {datetime.now().isoformat()}"
    self.progress_thread.cancel.emit()
    self.update_changed_status(UploadStatusValues.Cancelled.name)

  def update_changed_status(self, status: str = UploadStatusValues.Queued.name) -> None:
    """
    Updates the status of the upload task.

    Explanation:
        This function updates the status of the upload task to the specified status value.
        It then updates the model document in the database and emits a signal for the status change.

    Args:
        status (str): The status value to update the upload task to. Defaults to 'Queued'.
    """
    if self.upload_model and self.db_api:
      self.upload_model.status = status
      self.db_api.update_model_document(self.upload_model)
    self.status_changed.emit(status)

  def check_if_cancelled(self) -> bool:
    """
    Checks if the data upload task has been cancelled and finalizes the upload if cancelled.

    Returns:
        bool: True if the task is cancelled, False otherwise.
    """
    if self.cancelled:
      self.update_log(
        "User cancelled the upload, hence finalizing the upload!",
        self.logger.info,
      )
      self.finalize_upload_task(UploadStatusValues.Cancelled.name)
    return self.cancelled

  def generate_eln_file(self, tmp_dir: str) -> str:
    """
    Generates an ELN file based on the configuration model and project upload items.

    Args:
        tmp_dir (str): The temporary directory path to store the ELN file.

    Returns:
        str: The path to the generated ELN file.
        If an error occurs during export, an empty string is returned.
    """
    self.update_log(f"Step 1: Generating ELN file for project: {self.project_name}", self.logger.info)
    if self.config_model is None or self.config_model.project_upload_items is None:
      self.update_log("Config model or project_upload_items is not set!", self.logger.error)
      return ""
    eln_file = join(tmp_dir, f"{''.join(x for x in self.project_name if x.isalnum())}.eln")
    Path(eln_file).touch()
    data_types = [
      data_type.lower()
      for data_type, value in self.config_model.project_upload_items.items()
      if value
    ]
    try:
      exportELN(self.backend, [self.project_doc_id], eln_file, data_types)
    except Exception as e:
      self.update_log(f"Error while exporting ELN file for project:"
                      f" {self.project_name}, error: {e}", self.logger.error)
      eln_file = ""
    return eln_file



""" PYTHON MIXIN FOR BACKEND containing all the functions that output to CLI """
# mypy: ignore-errors
from .miscTools import createDirName

class CLI_Mixin:
  """ Python Mixin for backend containing all the functions that output to CLI """

  def output(self, docType, printID=False, **kwargs):
    """
    output view to screen
    - length of output 100 character

    Args:
      docType (string): document type to output
      printID (bool):  include docID in output string
      **kwargs (dict): additional parameter

    Returns:
        string: output incl. \n
    """
    outString = []
    widthArray = [25,25,25,25]
    for idx,item in enumerate(i for group in self.db.dataHierarchy[docType]['meta']
                              for i in self.db.dataHierarchy[docType]['meta'][group]):
      width = widthArray[idx] if idx<len(widthArray) else 0
      if width!=0:
        formatString = '{0: <'+str(abs(width))+'}'
        outString.append(formatString.format(item['name'].replace('-','')) )
    outString = '|'.join(outString)+'\n'
    outString += '-'*104+'\n'
    dataList  = self.db.getView(f'viewDocType/{docType}')
    for lineNum, lineItem in enumerate(dataList):
      rowString = []
      if lineNum>20:
        outString += f'... continued for {len(dataList)} items in total\n'
        break
      for idx, item in enumerate(i for group in self.db.dataHierarchy[docType]['meta']
                                 for i in self.db.dataHierarchy[docType]['meta'][group]):
        width = widthArray[idx] if idx<len(widthArray) else 0
        if width!=0:
          formatString = '{0: <'+str(abs(width))+'}'
          if isinstance(lineItem['value'][idx], str ):
            contentString = lineItem['value'][idx]
          elif isinstance(lineItem['value'][idx], (bool,int)):
            contentString = str(lineItem['value'][idx])
          elif lineItem['value'][idx] is None:
            contentString = '--'
          else: #list
            contentString = ' '.join(lineItem['value'][idx])
          contentString = contentString.replace('\n',' ')
          if width<0:  #test if value as non-trivial length
            if lineItem['value'][idx] in ['true', 'false']:
              contentString = lineItem['value'][idx]
            elif isinstance(lineItem['value'][idx], bool ) or lineItem['value'][idx] is None:
              contentString = str(lineItem['value'][idx])
            elif len(lineItem['value'][idx])>1 and len(lineItem['value'][idx][0])>3:
              contentString = 'true'
            else:
              contentString = 'false'
                    # contentString = True if contentString=='true' else False
          rowString.append(formatString.format(contentString)[:abs(width)] )
      if printID:
        rowString.append(' '+lineItem['id'])
      outString += '|'.join(rowString)+'\n'
    return outString


  def outputTags(self, tag='', **kwargs):
    """
    output view to screen
    - length of output 100 character

    Args:
      tag (string): tag to be listed, if empty: print all
      **kwargs (dict): additional parameter

    Returns:
        string: output incl. \n
    """
    outString = [f'{0: <10}'.format('Tags'), f'{0: <60}'.format('Name'), f'{0: <10}'.format('ID')]
    outString = '|'.join(outString)+'\n'
    outString += '-'*106+'\n'
    view = None
    if tag=='':
      view = self.db.getView('viewIdentify/viewTags')
    else:
      view = self.db.getView('viewIdentify/viewTags', preciseKey=f'#{tag}')
    for lineItem in view:
      rowString = [f'{0: <10}'.format(lineItem['key']), f'{0: <60}'.format(lineItem['value']), f'{0: <10}'.format(lineItem['id'])]
      outString += '|'.join(rowString)+'\n'
    return outString


  def outputHierarchy(self, onlyHierarchy=True, addID=False, addTags=None, **kwargs):
    """
    output hierarchical structure in database
    - convert view into native dictionary
    - ignore key since it is always the same

    Args:
       onlyHierarchy (bool): only print project,steps,tasks or print all (incl. measurements...)[default print all]
       addID (bool): add docID to output
       addTags (string): add tags, comments, objective to output ['all','tags',None]
       **kwargs (dict): additional parameter, i.e. callback

    Returns:
        string: output incl. \n
    """
    from anytree import PreOrderIter
    if len(self.hierStack) == 0:
      return 'Warning: pasta.outputHierarchy No project selected'
    hierString = ' '.join(self.hierStack)
    hierarchy = self.db.getHierarchy(hierString)
    output = ""
    for node in PreOrderIter(hierarchy):
      output +='  '*node.depth+node.name+' | '+'/'.join(node.docType)+' | '+node.id+'\n'
    return output


  def getEditString(self):
    """
    Return org-mode markdown string of hierarchy tree
      complicated style: this document and all its children and grandchildren...

    Returns:
        string: output incl. \n
    """
    return self.outputHierarchy(True,True,'tags')


  def outputQR(self):
    """
    output list of sample qr-codes

    Returns:
        string: output incl. \n
    """
    outString = f"{'QR': <36}|{'Name': <36}|{'ID': <36}\n"
    outString += '-'*110+'\n'
    for item in self.db.getView('viewIdentify/viewQR'):
      outString += f"{item['key'][:36]: <36}|{item['value'][:36]: <36}|{item['id'][:36]: <36}\n"
    return outString


  def outputSHAsum(self):
    """
    output list of measurement SHA-sums of files

    Returns:
        string: output incl. \n
    """
    outString = f"{'SHAsum': <32}|{'Name': <40}|{'ID': <25}\n"
    outString += '-'*110+'\n'
    for item in self.db.getView('viewIdentify/viewSHAsum'):
      key = item['key'] or '-empty-'
      outString += f"{key[:32]: <32}|{item['value'][:40]: <40}|{item['id']: <25}\n"
    return outString



# pylint: skip-file
# mypy: ignore-errors
from __future__ import annotations

import base64
import enum
import hashlib
import os
import secrets
import time
from dataclasses import dataclass
from pathlib import Path
from typing import (
    BinaryIO,
    Optional,
    Union,
)

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.kdf import scrypt

from .exceptions import (
    Error,
    ParseError,
    VerifyError,
)
from .helpers import (
    Reader,
    check_comment,
    read_data,
)

ALG_LEN = 2
KDF_PARAM_LEN = 8
KEY_ID_LEN = 8
KEY_LEN = 32
SALT_LEN = 32
CHECKSUM_LEN = 32
SIG_LEN = 64

KEYNUM_PK_LEN = KEY_ID_LEN + KEY_LEN
KEYNUM_SK_LEN = KEY_ID_LEN + (KEY_LEN << 1) + CHECKSUM_LEN

OPSLIMIT = 1_048_576
MEMLIMIT = 33_554_432
MEMLIMIT_MAX = 1_073_741_824
N_LOG2_MAX = 20

SIG_EXT = 'minisig'
BYTE_ORDER = 'little'
DEFAULT_SK_PATH = '~/.minisign/minisign.key'

UNTRUSTED_COMMENT_PREFIX = 'untrusted comment: '
TRUSTED_COMMENT_PREFIX = 'trusted comment: '
TRUSTED_COMMENT_PREFIX_LEN = len(TRUSTED_COMMENT_PREFIX)


@enum.unique
class SignatureAlgorithm(bytes, enum.Enum):
    PURE_ED_DSA = bytes([0x45, 0x64])
    PREHASHED_ED_DSA = bytes([0x45, 0x44])


@enum.unique
class KDFAlgorithm(bytes, enum.Enum):
    SCRYPT = bytes([0x53, 0x63])


@enum.unique
class CksumAlgorithm(bytes, enum.Enum):
    BLAKE2b = bytes([0x42, 0x32])


@dataclass(frozen=True)
class Signature:
    _untrusted_comment: str
    _signature_algorithm: SignatureAlgorithm
    _key_id: bytes
    _signature: bytes
    _trusted_comment: str
    _global_signature: bytes

    @classmethod
    def from_bytes(cls, data: bytes) -> Signature:
        lines = data.splitlines()
        if len(lines) < 4:
            raise ParseError('incomplete encoded signature')
        glob_sig = base64.standard_b64decode(lines[3])
        if len(glob_sig) != SIG_LEN:
            raise ParseError('invalid encoded signature')
        buf = Reader(base64.standard_b64decode(lines[1]))
        return cls(
            _untrusted_comment=lines[0].decode(),
            _signature_algorithm=SignatureAlgorithm(buf.read(ALG_LEN)),
            _key_id=buf.read(KEY_ID_LEN),
            _signature=buf.read(SIG_LEN),
            _trusted_comment=lines[2].decode(),
            _global_signature=glob_sig,
        )

    @classmethod
    def from_file(cls, path: Union[str, os.PathLike]) -> Signature:
        with open(path, 'rb') as f:
            return cls.from_bytes(f.read())

    @property
    def untrusted_comment(self) -> str:
        return self._untrusted_comment

    def set_untrusted_comment(self, value: str):
        check_comment(value)
        self.__dict__['_untrusted_comment'] = value

    @property
    def trusted_comment(self) -> str:
        return self._trusted_comment[TRUSTED_COMMENT_PREFIX_LEN:]

    def __bytes__(self) -> bytes:
        return b'\n'.join((
            self._untrusted_comment.encode(),
            base64.standard_b64encode(
                self._signature_algorithm.value +
                self._key_id +
                self._signature
            ),
            self._trusted_comment.encode(),
            base64.standard_b64encode(self._global_signature),
        ))

    def _is_prehashed(self) -> bool:
        return self._signature_algorithm == SignatureAlgorithm.PREHASHED_ED_DSA


@dataclass(frozen=True)
class KeynumPK:
    key_id: bytes
    public_key: bytes

    @classmethod
    def from_bytes(cls, data: Union[bytes, Reader]) -> KeynumPK:
        assert len(data) == KEYNUM_PK_LEN
        if isinstance(data, bytes):
            data = Reader(data)
        return cls(
            key_id=data.read(KEY_ID_LEN),
            public_key=data.read(KEY_LEN),
        )

    def __bytes__(self) -> bytes:
        return self.key_id + self.public_key


@dataclass(frozen=True)
class PublicKey:
    _untrusted_comment: Optional[str]
    _signature_algorithm: SignatureAlgorithm
    _keynum_pk: KeynumPK

    @classmethod
    def from_base64(cls, s: Union[bytes, str]) -> PublicKey:
        buf = Reader(base64.standard_b64decode(s))
        return cls(
            _untrusted_comment=None,
            _signature_algorithm=SignatureAlgorithm(buf.read(ALG_LEN)),
            _keynum_pk=KeynumPK.from_bytes(buf),
        )

    @classmethod
    def from_bytes(cls, data: bytes) -> PublicKey:
        lines = data.splitlines()
        if len(lines) < 2:
            raise ParseError('incomplete encoded public key')
        pk = cls.from_base64(lines[1])
        pk.set_untrusted_comment(lines[0].decode())
        return pk

    @classmethod
    def from_file(cls, path: Union[str, os.PathLike]) -> PublicKey:
        with open(path, 'rb') as f:
            return cls.from_bytes(f.read())

    @classmethod
    def from_secret_key(cls, secret_key: SecretKey) -> PublicKey:
        key_id = bytes(secret_key._keynum_sk.key_id)
        return cls(
            _untrusted_comment=f'{UNTRUSTED_COMMENT_PREFIX}'
                               f'minisign public key '
                               f'{key_id.hex().upper()}',
            _signature_algorithm=secret_key._signature_algorithm,
            _keynum_pk=KeynumPK(
                key_id=key_id,
                public_key=bytes(secret_key._keynum_sk.public_key),
            ),
        )

    @property
    def untrusted_comment(self) -> Optional[str]:
        return self._untrusted_comment

    def set_untrusted_comment(self, value: Optional[str]):
        check_comment(value)
        self.__dict__['_untrusted_comment'] = value

    def verify(self, data: Union[bytes, BinaryIO], signature: Signature):
        if self._keynum_pk.key_id != signature._key_id:
            raise VerifyError('incompatible key identifiers')
        if not signature._trusted_comment.startswith(TRUSTED_COMMENT_PREFIX):
            raise VerifyError('unexpected format for the trusted comment')
        pk = ed25519.Ed25519PublicKey.from_public_bytes(
            self._keynum_pk.public_key)
        try:
            pk.verify(
                signature._signature,
                read_data(data, signature._is_prehashed()),
            )
            pk.verify(
                signature._global_signature,
                signature._signature + signature.trusted_comment.encode(),
            )
        except InvalidSignature as err:
            raise VerifyError(err)

    def verify_file(
        self,
        path: Union[str, os.PathLike],
        signature: Optional[Signature] = None,
    ):
        if signature is None:
            signature = Signature.from_file(f'{path}.{SIG_EXT}')
        with open(path, 'rb') as f:
            self.verify(f, signature)

    def to_base64(self) -> bytes:
        return base64.standard_b64encode(
            self._signature_algorithm.value +
            bytes(self._keynum_pk)
        )

    def __bytes__(self) -> bytes:
        return b'\n'.join((
            (
                f'{UNTRUSTED_COMMENT_PREFIX}minisign public key '
                f'{self._keynum_pk.key_id.hex().upper()}'
                if self._untrusted_comment is None else
                self._untrusted_comment
            ).encode(),
            self.to_base64(),
        ))


@dataclass(frozen=True, repr=False)
class KeynumSK:
    key_id: bytearray
    secret_key: bytearray
    public_key: bytearray
    checksum: bytearray

    @classmethod
    def from_bytes(cls, data: Union[bytes, Reader]) -> KeynumSK:
        assert len(data) == KEYNUM_SK_LEN
        if isinstance(data, bytes):
            data = Reader(data)
        return cls(
            key_id=bytearray(data.read(KEY_ID_LEN)),
            secret_key=bytearray(data.read(KEY_LEN)),
            public_key=bytearray(data.read(KEY_LEN)),
            checksum=bytearray(data.read(CHECKSUM_LEN)),
        )

    def xor(self, key: bytes):
        assert len(key) == KEYNUM_SK_LEN
        buf = Reader(key)
        for (l, size) in (
            (self.key_id, KEY_ID_LEN),
            (self.secret_key, KEY_LEN),
            (self.public_key, KEY_LEN),
            (self.checksum, CHECKSUM_LEN),
        ):
            for idx, (v1, v2) in enumerate(zip(l[:], buf.read(size))):
                l[idx] = v1 ^ v2

    def __bytes__(self) -> bytes:
        return (
            bytes(self.key_id) +
            bytes(self.secret_key) +
            bytes(self.public_key) +
            bytes(self.checksum)
        )


@dataclass(frozen=True, repr=False)
class SecretKey:
    _untrusted_comment: str
    _signature_algorithm: SignatureAlgorithm
    _kdf_algorithm: KDFAlgorithm
    _cksum_algorithm: CksumAlgorithm
    _kdf_salt: bytes
    _kdf_opslimit: int
    _kdf_memlimit: int
    _keynum_sk: KeynumSK

    @classmethod
    def from_bytes(cls, data: bytes) -> SecretKey:
        lines = data.splitlines()
        if len(lines) < 2:
            raise ParseError('incomplete encoded secret key')
        buf = Reader(base64.standard_b64decode(lines[1]))
        return cls(
            _untrusted_comment=lines[0].decode(),
            _signature_algorithm=SignatureAlgorithm(buf.read(ALG_LEN)),
            _kdf_algorithm=KDFAlgorithm(buf.read(ALG_LEN)),
            _cksum_algorithm=CksumAlgorithm(buf.read(ALG_LEN)),
            _kdf_salt=buf.read(SALT_LEN),
            _kdf_opslimit=int.from_bytes(buf.read(KDF_PARAM_LEN), BYTE_ORDER),
            _kdf_memlimit=int.from_bytes(buf.read(KDF_PARAM_LEN), BYTE_ORDER),
            _keynum_sk=KeynumSK.from_bytes(buf),
        )

    @classmethod
    def from_file(
        cls,
        path: Optional[Union[str, os.PathLike]] = None,
    ) -> SecretKey:
        if path is None:
            path = Path(DEFAULT_SK_PATH).expanduser().resolve(strict=True)
        with open(path, 'rb') as f:
            return cls.from_bytes(f.read())

    @property
    def untrusted_comment(self) -> str:
        return self._untrusted_comment

    def set_untrusted_comment(self, value: str):
        check_comment(value)
        self.__dict__['_untrusted_comment'] = value

    def get_public_key(self) -> PublicKey:
        return PublicKey.from_secret_key(self)

    def decrypt(self, password: str):
        self._crypt(password)
        if self._calc_checksum() != bytes(self._keynum_sk.checksum):
            raise Error('wrong password for that key')

    def encrypt(self, password: str):
        self._crypt(password)

    def _crypt(self, password: str):
        if self._kdf_memlimit > MEMLIMIT_MAX:
            raise Error('memlimit too high')
        opslimit = max(32768, self._kdf_opslimit)
        n_log2 = 1
        r = 8
        p = 0
        if opslimit < self._kdf_memlimit // 32:
            maxn = opslimit // (r * 4)
            p = 1
        else:
            maxn = self._kdf_memlimit // (r * 128)
        while n_log2 < 63:
            if 1 << n_log2 > maxn // 2:
                break
            n_log2 += 1
        if not p:
            p = min(0x3fffffff, (opslimit // 4) // (1 << n_log2)) // r
        if n_log2 > N_LOG2_MAX:
            raise Error('n_log2 too high')
        self._keynum_sk.xor(scrypt.Scrypt(
            salt=self._kdf_salt,
            length=KEYNUM_SK_LEN,
            n=1 << n_log2,
            r=r,
            p=p,
        ).derive(password.encode()))

    def sign(
        self,
        data: Union[bytes, BinaryIO],
        *,
        prehash: bool = True,
        untrusted_comment: Optional[str] = None,
        trusted_comment: Optional[str] = None,
    ) -> Signature:
        untrusted_comment = (
            f'{UNTRUSTED_COMMENT_PREFIX}minisign signature '
            f'{self._keynum_sk.key_id.hex().upper()}'
            if untrusted_comment is None else
            untrusted_comment
        )
        check_comment(untrusted_comment)
        trusted_comment = (
            f'timestamp:{int(time.time())}'
            if trusted_comment is None else
            trusted_comment
        )
        check_comment(trusted_comment)
        pk = ed25519.Ed25519PrivateKey.from_private_bytes(
            self._keynum_sk.secret_key)
        sig_sig = pk.sign(read_data(data, prehash))
        return Signature(
            _untrusted_comment=untrusted_comment,
            _signature_algorithm=(
                SignatureAlgorithm.PREHASHED_ED_DSA
                if prehash else
                SignatureAlgorithm.PURE_ED_DSA
            ),
            _key_id=self._keynum_sk.key_id,
            _signature=sig_sig,
            _trusted_comment=f'{TRUSTED_COMMENT_PREFIX}{trusted_comment}',
            _global_signature=pk.sign(sig_sig + trusted_comment.encode()),
        )

    def sign_file(
        self,
        path: Union[str, os.PathLike],
        *,
        prehash: bool = False,
        untrusted_comment: Optional[str] = None,
        trusted_comment: Optional[str] = None,
        drop_signature: bool = False,
    ) -> Signature:
        with open(path, 'rb') as f:
            sig = self.sign(
                f,
                prehash=prehash,
                untrusted_comment=untrusted_comment,
                trusted_comment=trusted_comment,
            )
        if drop_signature:
            with open(f'{path}.{SIG_EXT}', 'wb') as f1:
                f1.write(bytes(sig))
                f1.write(b'\n')
        return sig

    def _calc_checksum(self) -> bytes:
        hasher = hashlib.blake2b(digest_size=CHECKSUM_LEN)
        hasher.update(self._signature_algorithm.value)
        hasher.update(self._keynum_sk.key_id)
        hasher.update(self._keynum_sk.secret_key)
        hasher.update(self._keynum_sk.public_key)
        return hasher.digest()

    def _update_checksum(self):
        self._keynum_sk.checksum[0:] = self._calc_checksum()

    def __bytes__(self) -> bytes:
        return b'\n'.join((
            self._untrusted_comment.encode(),
            base64.standard_b64encode(
                self._signature_algorithm.value +
                self._kdf_algorithm.value +
                self._cksum_algorithm.value +
                self._kdf_salt +
                self._kdf_opslimit.to_bytes(KDF_PARAM_LEN, BYTE_ORDER) +
                self._kdf_memlimit.to_bytes(KDF_PARAM_LEN, BYTE_ORDER) +
                bytes(self._keynum_sk)
            ),
        ))


@dataclass(frozen=True, repr=False)
class KeyPair:
    secret_key: SecretKey
    public_key: PublicKey

    @classmethod
    def generate(cls) -> KeyPair:
        private_key = ed25519.Ed25519PrivateKey.generate()
        key_id = secrets.token_bytes(KEY_ID_LEN)
        sk = SecretKey(
            _untrusted_comment=f'{UNTRUSTED_COMMENT_PREFIX}'
                               f'minisign secret key '
                               f'{key_id.hex().upper()}',
            _signature_algorithm=SignatureAlgorithm.PURE_ED_DSA,
            _kdf_algorithm=KDFAlgorithm.SCRYPT,
            _cksum_algorithm=CksumAlgorithm.BLAKE2b,
            _kdf_salt=secrets.token_bytes(SALT_LEN),
            _kdf_opslimit=OPSLIMIT,
            _kdf_memlimit=MEMLIMIT,
            _keynum_sk=KeynumSK(
                key_id=bytearray(key_id),
                secret_key=bytearray(private_key.private_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PrivateFormat.Raw,
                    encryption_algorithm=serialization.NoEncryption(),
                )),
                public_key=bytearray(private_key.public_key().public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw,
                )),
                checksum=bytearray(CHECKSUM_LEN),
            ),
        )
        sk._update_checksum()
        return cls(secret_key=sk, public_key=PublicKey.from_secret_key(sk))



""" widget that shows the table of the items """
import itertools
import re, logging
from enum import Enum
from pathlib import Path
from typing import Any
from PySide6.QtWidgets import QWidget, QVBoxLayout, QTableView, QFileDialog, QMessageBox, QHeaderView, QLineEdit, QComboBox, QMenu # pylint: disable=no-name-in-module
from PySide6.QtCore import Qt, Slot, QSortFilterProxyModel, QModelIndex       # pylint: disable=no-name-in-module
from PySide6.QtGui import QStandardItemModel, QStandardItem            # pylint: disable=no-name-in-module
from .tableHeader import TableHeader
from ..guiStyle import IconButton, Action, TextButton, Label, widgetAndLayout, widgetAndLayoutGrid, space
from ..guiCommunicate import Communicate

#Scan button to more button
class Table(QWidget):
  """ widget that shows the table of the items """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
    """
    super().__init__()
    self.comm = comm
    comm.changeTable.connect(self.change)
    comm.stopSequentialEdit.connect(self.stopSequentialEditFunction)
    self.stopSequentialEdit = False
    self.data:list[dict[str,Any]] = []
    self.models:list[QSortFilterProxyModel] = []
    self.docType = ''
    self.projID = ''
    self.filterHeader:list[str] = []
    self.showAll= True

    ### GUI elements
    mainL = QVBoxLayout()
    mainL.setSpacing(0)
    mainL.setContentsMargins(space['s'], space['s'], space['s'], space['s'])
    # header
    self.headerW, headerL = widgetAndLayout('H', mainL, 'm')
    self.headerW.hide()
    self.headline = Label('','h1', headerL)
    self.showState= Label('', 'h3', headerL)
    headerL.addStretch(1)
    self.addBtn = TextButton('Add',  self, [Command.ADD_ITEM],   headerL)

    self.selectionBtn = TextButton('Selection', self, [], headerL)
    selectionMenu = QMenu(self)
    Action('Toggle selection',self, [Command.TOGGLE_SELECTION], selectionMenu)
    selectionMenu.addSeparator()
    Action('Group Edit',      self, [Command.GROUP_EDIT],       selectionMenu)
    Action('Sequential edit', self, [Command.SEQUENTIAL_EDIT],  selectionMenu)
    Action('Rerun extractors',self, [Command.RERUN_EXTRACTORS], selectionMenu)
    Action('Delete',          self, [Command.DELETE],           selectionMenu)
    self.selectionBtn.setMenu(selectionMenu)

    self.visibilityBtn = TextButton('Visibility', self, [], headerL)
    visibilityMenu = QMenu(self)
    Action(                    'Add Filter',                       self, [Command.ADD_FILTER],  visibilityMenu)
    self.showHidden   = Action('Show/hide hidden ___',             self, [Command.SHOW_ALL],    visibilityMenu)
    self.toggleHidden = Action('Invert hidden status of selected', self, [Command.TOGGLE_HIDE], visibilityMenu)
    self.visibilityBtn.setMenu(visibilityMenu)

    more = TextButton('More', self, [], headerL)
    self.moreMenu = QMenu(self)
    Action('Export to csv',            self, [Command.EXPORT],   self.moreMenu)
    self.actionChangeColums = Action('Change columns',  self, [Command.CHANGE_COLUMNS], self.moreMenu)  #add this action at end
    more.setMenu(self.moreMenu)

    # filter
    _, self.filterL = widgetAndLayoutGrid(mainL, top='s', bottom='s')
    # table
    self.table = QTableView(self)
    self.table.verticalHeader().hide()
    self.table.clicked.connect(self.cellClicked)
    self.table.doubleClicked.connect(self.cell2Clicked)
    self.table.setSortingEnabled(True)
    self.table.setAlternatingRowColors(True)
    header = self.table.horizontalHeader()
    header.setSectionsMovable(True)
    header.setSortIndicatorShown(True)
    header.setMaximumSectionSize(self.comm.backend.configuration['GUI']['maxTableColumnWidth'])
    header.resizeSections(QHeaderView.ResizeMode.ResizeToContents)
    header.setStretchLastSection(True)
    mainL.addWidget(self.table)
    self.setLayout(mainL)


  @Slot(str, str)
  def change(self, docType:str, projID:str) -> None:
    """
    What happens when the table changes its raw information

    Args:
      docType (str): document type; leave empty for redraw
      projID (str): id of project
    """
    if docType!=self.docType or projID!=self.projID:
      logging.debug('table:changeTable |%s|%s|',docType, projID)
    self.models = []
    #if not docType:  #only remove old filters, when docType changes
    #   make sure internal updates are accounted for: i.e. comment
    for i in reversed(range(self.filterL.count())):
      self.filterL.itemAt(i).widget().setParent(None)
    if docType!='':
      self.docType = docType
      self.projID  = projID
    if self.docType=='_tags_':
      self.addBtn.hide()
      if self.showAll:
        self.data = self.comm.backend.db.getView('viewIdentify/viewTagsAll')
      else:
        self.data = self.comm.backend.db.getView('viewIdentify/viewTags')
      self.filterHeader = ['tag','name','type']
      self.headline.setText('TAGS')
      self.actionChangeColums.setVisible(False)
    else:
      self.addBtn.show()
      if self.docType.startswith('x0'):
        self.selectionBtn.hide()
        self.toggleHidden.setVisible(False)
      else:
        self.selectionBtn.show()
        self.toggleHidden.setVisible(True)
      path = (f'viewDocType/{self.docType}All' if self.showAll else f'viewDocType/{self.docType}')
      if self.projID=='':
        self.data = self.comm.backend.db.getView(path)
      else:
        self.data = self.comm.backend.db.getView(path, preciseKey=self.projID)
      # filter multiple lines of the same item: #https://stackoverflow.com/questions/11092511/list-of-unique-dictionaries
      self.data = list({v['id']:v for v in self.data}.values())
      self.showState.setText('(show all rows)' if self.showAll else '(hide hidden rows)')
      docLabel = 'Unidentified'
      if self.docType=='-':
        self.actionChangeColums.setVisible(False)
      else:
        self.actionChangeColums.setVisible(True)
        if self.docType in self.comm.backend.db.dataLabels:
          docLabel = self.comm.backend.db.dataLabels[self.docType]
      if self.projID:
        self.headline.setText(docLabel)
        self.showHidden.setText(f'Show/hide hidden {docLabel.lower()}')
      else:
        self.comm.changeSidebar.emit('')  #close the project in sidebar
        self.headline.setText(f'All {docLabel.lower()}')
        self.showHidden.setText(f'Show/hide all hidden {docLabel.lower()}')
      self.filterHeader = self.comm.backend.db.getColumnNames()[self.docType].split(',')
      self.filterHeader = [i[1:] if i[0]=='-'   else i for i in self.filterHeader]  #change -something to something
      self.filterHeader = [i[2:] if i[:2]=='#_' else i for i in self.filterHeader]  #change #_something to something
    self.headerW.show()
    nrows, ncols = len(self.data), len(self.filterHeader)
    model = QStandardItemModel(nrows, ncols)
    model.setHorizontalHeaderLabels(self.filterHeader)
    for i, j in itertools.product(range(nrows), range(ncols)):
      if self.docType=='_tags_':  #tags list
        if j==0:
          if self.data[i]['key']=='_curated':
            item = QStandardItem('_curated_')
          elif re.match(r'_\d', self.data[i]['key']):
            item = QStandardItem('\u2605'*int(self.data[i]['key'][1]))
          else:
            item = QStandardItem(self.data[i]['key'])
        else:
          item = QStandardItem(self.data[i]['value'][j-1])
      elif self.data[i]['value'][j] is None or not self.data[i]['value'][j]:  #None, False
        item = QStandardItem('-') # if you want to add nice glyphs, see also below \u00D7')
        # item.setFont(QFont("Helvetica [Cronyx]", 16))
      elif isinstance(self.data[i]['value'][j], bool): #True
        item = QStandardItem('Y') # \u2713')
        # item.setFont(QFont("Helvetica [Cronyx]", 16))
      elif isinstance(self.data[i]['value'][j], list):                  #list, e.g. qrCodes
        item = (QStandardItem(', '.join(self.data[i]['value'][j])) if isinstance(
            self.data[i]['value'][j][0], str) else QStandardItem(', '.join(
                [str(i) for i in self.data[i]['value'][j]])))
      elif isinstance(self.data[i]['value'][j], str) and re.match(r'^[a-z\-]-[a-z0-9]{32}$',self.data[i]['value'][j]):      #Link
        item = QStandardItem('oo') # \u260D')
        # item.setFont(QFont("Helvetica [Cronyx]", 16))
      else:
        if self.filterHeader[j]=='tags':
          tags = self.data[i]['value'][j].split(' ')
          if '_curated' in tags:
            tags[tags.index('_curated')] = '_curated_'
          for iStar in range(1,6):
            if f'_{str(iStar)}' in tags:
              tags[tags.index(f'_{str(iStar)}')] = '*'*iStar
          text = ' '.join(tags)
        else:
          text = str(self.data[i]['value'][j])  #could be a number
        item = QStandardItem(text)
      if j==0:
        doc = self.comm.backend.db.getDoc(self.data[i]['id'])
        if [b for b in doc['-branch'] if False in b['show']]:
          item.setText( item.text()+'  \U0001F441' )
        item.setAccessibleText(doc['_id'])
        if self.docType!='x0':
          item.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled)
          item.setCheckState(Qt.CheckState.Unchecked)
      else:
        item.setFlags(Qt.ItemFlag.ItemIsEnabled)
      model.setItem(i, j, item)
    self.models.append(model)                                                                                # type: ignore[arg-type]
    self.table.setModel(self.models[-1])
    self.table.show()
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.ADD_ITEM:
      self.comm.formDoc.emit({'-type':[self.docType], '_projectID':self.projID})
      self.comm.changeTable.emit(self.docType, self.projID)
      if self.docType=='x0':
        self.comm.changeSidebar.emit('redraw')
    elif command[0] is Command.ADD_FILTER:
      # gui
      _, rowL = widgetAndLayout('H', self.filterL, 'm', 'xl', '0', 'xl')
      text = QLineEdit('')
      rowL.addWidget(text)
      select = QComboBox()
      select.addItems(self.filterHeader)
      select.currentIndexChanged.connect(self.filterChoice)
      select.setMinimumWidth(max(len(i) for i in self.filterHeader)*14)
      select.setAccessibleName(str(len(self.models)))
      rowL.addWidget(select)
      IconButton('fa5s.minus-square', self, [Command.DELETE_FILTER, len(self.models)], rowL)
      # data
      filterModel = QSortFilterProxyModel()
      text.textChanged.connect(filterModel.setFilterRegularExpression)
      filterModel.setSourceModel(self.models[-1])
      filterModel.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
      filterModel.setFilterKeyColumn(0)
      self.models.append(filterModel)
      self.table.setModel(self.models[-1])
    elif command[0] is Command.GROUP_EDIT:
      intersection = None
      docIDs = []
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          docIDs.append(docID)
          thisKeys = set(self.comm.backend.db.getDoc(docID))
          if intersection is None:
            intersection = thisKeys
          else:
            intersection = intersection.intersection(thisKeys)
      #remove keys that should not be group edited and build dict
      if intersection is not None:
        intersection = intersection.difference({'-branch', '-user', '-client', 'metaVendor', 'shasum', \
            '_id', 'metaUser', '_rev', '-name', '-date', 'image', '_attachments','links'})
        intersectionDict:dict[str,Any] = {i:'' for i in intersection}
        intersectionDict['-tags'] = []
        intersectionDict['-type'] = [self.docType]
        intersectionDict['_ids'] = docIDs
        self.comm.formDoc.emit(intersectionDict)
        self.comm.changeDetails.emit('redraw')
        self.comm.changeTable.emit(self.docType, self.projID)
    elif command[0] is Command.SEQUENTIAL_EDIT:
      self.stopSequentialEdit = False
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          self.comm.formDoc.emit(self.comm.backend.db.getDoc(docID))
        if self.stopSequentialEdit:
          break
      self.comm.changeTable.emit(self.docType, self.projID)
    elif command[0] is Command.DELETE:
      ret = None
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          if ret is None:
            ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete the data?',
                QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)
          if ret==QMessageBox.StandardButton.Yes:
            doc = self.comm.backend.db.getDoc(docID)
            for branch in doc['-branch']:
              if branch['path'] is not None:
                oldPath = self.comm.backend.basePath/branch['path']
                if oldPath.exists():
                  newPath = oldPath.parent / f'trash_{oldPath.name}'
                  oldPath.rename(newPath)
            self.comm.backend.db.remove(docID)
      self.comm.changeTable.emit(self.docType, self.projID)
    elif command[0] is Command.CHANGE_COLUMNS:
      dialog = TableHeader(self.comm, self.docType)
      dialog.exec()
    elif command[0] is Command.EXPORT:
      fileName = QFileDialog.getSaveFileName(self,'Export to ..',str(Path.home()),'*.csv')[0]
      if not fileName.endswith('.csv'):
        fileName += '.csv'
      with open(fileName,'w', encoding='utf-8') as fOut:
        header = [f'"{i}"' for i in self.filterHeader]
        fOut.write(','.join(header)+'\n')
        for row in range(self.models[-1].rowCount()):
          rowContent = []
          for col in range(self.models[-1].columnCount()):
            value = self.models[-1].index( row, col, QModelIndex() ).data( Qt.ItemDataRole.DisplayRole )
            rowContent.append(f'"{value}"')
          fOut.write(','.join(rowContent)+'\n')
    elif command[0] is Command.TOGGLE_HIDE:
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          self.comm.backend.db.hideShow(docID)
      if self.docType=='x0':
        self.comm.changeSidebar.emit('redraw')
      self.change('','')  # redraw table
    elif command[0] is Command.TOGGLE_SELECTION:
      for row in range(self.models[-1].rowCount()):
        item,_ = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          item.setCheckState(Qt.CheckState.Unchecked)
        else:
          item.setCheckState(Qt.CheckState.Checked)
    elif command[0] is Command.SHOW_ALL:
      self.showAll = not self.showAll
      self.change('','')  # redraw table
    elif command[0] is Command.RERUN_EXTRACTORS:
      redraw = False
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          doc = self.comm.backend.db.getDoc(docID)
          if doc['-branch'][0]['path'] is None:
            continue
          redraw = True
          oldDocType = doc['-type']
          doc['-type'] = ['']
          if doc['-branch'][0]['path'].startswith('http'):
            path = Path(doc['-branch'][0]['path'])
          else:
            path = self.comm.backend.basePath/doc['-branch'][0]['path']
          self.comm.backend.useExtractors(path, doc.get('shasum',''), doc)
          if doc['-type'][0] == oldDocType[0]:
            del doc['-branch']  #don't update
            self.comm.backend.db.updateDoc(doc, self.data[row]['id'])
          else:
            self.comm.backend.db.remove( self.data[row]['id'] )
            del doc['_id']
            del doc['_rev']
            doc['-name'] = doc['-branch'][0]['path']
            self.comm.backend.addData('/'.join(doc['-type']), doc, doc['-branch'][0]['stack'])
      if redraw:
        self.change('','')  # redraw table
        self.comm.changeDetails.emit('redraw')
    elif command[0] is Command.DELETE_FILTER: # Remove filter from list of filters
      row = command[1]
      #print('Delete filter row', row)
      for i in range(row, self.filterL.count()):        #e.g. model 1 is in row=0, so start in 1 for renumbering
        minusBtnW = self.filterL.itemAt(i).widget().layout().itemAt(2).widget()
        minusBtnW.setAccessibleName( str(int(minusBtnW.accessibleName())-1) )  #rename: -1 from accessibleName
      del self.models[row]
      self.filterL.itemAt(row-1).widget().setParent(None) # e.g. model 1 is in row=0 for deletion
      for i in range(1, len(self.models)):
        self.models[i].setSourceModel(self.models[i-1])
      self.table.setModel(self.models[-1])
    else:
      print("**ERROR table menu unknown:",command)
    return


  def cellClicked(self, item:QStandardItem) -> None:
    """
    What happens when user clicks cell in table of tags, projects, samples, ...
    -> show details

    Args:
      item (QStandardItem): cell clicked
    """
    row = item.row()
    _, docID = self.itemFromRow(row)
    # column = item.column()
    if docID[0]=='x': #only show items for non-folders
      doc = self.comm.backend.db.getDoc(docID)
      if doc['-type'][0]=='x0':
        self.comm.changeProject.emit(docID,'')
        self.comm.changeSidebar.emit(docID)
      else:
        projID = doc['-branch'][0]['stack'][0]
        self.comm.changeProject.emit(projID, docID)
        self.comm.changeSidebar.emit(projID)
    else:
      self.comm.changeDetails.emit(docID)
    return


  def cell2Clicked(self, item:QStandardItem) -> None:
    """
    What happens when user double clicks cell in table of projects

    Args:
      item (QStandardItem): cell clicked
    """
    row = item.row()
    _, docID = self.itemFromRow(row)
    if self.docType=='x0':
      self.comm.changeProject.emit(docID, '')
      self.comm.changeSidebar.emit(docID)
    else:
      doc = self.comm.backend.db.getDoc(docID)
      self.comm.formDoc.emit(doc)
      self.comm.changeTable.emit('','')
      self.comm.changeDetails.emit('redraw')
    return


  @Slot()
  def stopSequentialEditFunction(self) -> None:
    """ Stop the sequential edit of a number of items """
    self.stopSequentialEdit=True
    return


  def itemFromRow(self, row:int) -> tuple[QStandardItem, str]:
    """
    get item from row by iterating through the proxyModels

    Args:
      row (int): row number

    Returns:
      QItem, str: the item and docID
    """
    index = self.models[-1].index(row,0)
    for idxModel in range(len(self.models)-1,0,-1):
      index = self.models[idxModel].mapToSource(index)
    item = self.models[0].itemFromIndex(index)                                                               # type: ignore[attr-defined]
    return item, item.accessibleText()


  def filterChoice(self, item:int) -> None:
    """
    Change the column which is used for filtering

    Args:
       item (int): column number to filter by
    """
    row = self.sender().accessibleName()                                                                     # type: ignore[attr-defined]
    self.models[int(row)].setFilterKeyColumn(item)
    return


class Command(Enum):
  """ Commands used in this file """
  ADD_FILTER       = 1
  ADD_ITEM         = 2
  TOGGLE_SELECTION = 3
  GROUP_EDIT       = 4
  SEQUENTIAL_EDIT  = 5
  TOGGLE_HIDE      = 6
  RERUN_EXTRACTORS = 7
  DELETE           = 8
  SHOW_ALL         = 9
  EXPORT           = 10
  CHANGE_COLUMNS   = 11
  DELETE_FILTER    = 12



#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: __init__.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.



"""extract data from vendor
- default png file
"""
import base64
from io import BytesIO
import numpy as np
from PIL import Image

def use(filePath, recipe='', saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    recipe (string): supplied to guide recipes
                     recipe is / separated hierarchical elements parent->child
    saveFileName (string): if given, save the image to this file-name

  Returns:
    dict: containing image, metaVendor, metaUser, recipe
  """
  # Extractor
  image = Image.open(filePath)
  metaVendor = image.info
  toDelete = []
  for key,value in metaVendor.items():
    if not isinstance(value, (str, int, float, list)):
      toDelete.append(key)
  for key in toDelete:
    del metaVendor[key]
  imgArr = np.array(image)
  if recipe == 'measurement/image/crop':    #: Crop 3/4 of the image
    newHeight = int(imgArr.shape[0]/2)
    newWidth  = int(imgArr.shape[1]/2)
    imgArr = imgArr[:newHeight, :newWidth, :]
  elif True or recipe == 'measurement/image': #: Default | uncropped
    recipe = 'measurement/image'
  metaUser   = {'number all pixel': int(np.prod(image.size[:2])) }

  #save to file
  imageData = Image.fromarray(imgArr)
  if saveFileName is not None:
    imageData.save(saveFileName)

  # convert PIL image to base64
  figfile = BytesIO()
  imageData.save(figfile, format="PNG")
  imageData = base64.b64encode(figfile.getvalue()).decode()
  imageData = "data:image/png;base64," + imageData

  # return everything
  return {'image':imageData, 'recipe':recipe, 'metaVendor':metaVendor, 'metaUser':metaUser}

  #other datatypes could follow here if statements are used
  #...
  #final return if nothing successful
  #return {'recipe': '-', 'metaVendor':{}, 'metaUser':{}}



""" Misc PRINTER functions that do not require instances """
import os, uuid
from typing import Any

def createQRcodeSheet(fileName:str="../qrCodes.pdf") -> None:
  """
  Documentation QR-codes
  - img = qrcode.make("testString",error_correction=qrcode.constants.ERROR_CORRECT_M)
  - or ERROR-CORRECT_H for better errorcorrection
  Sizes:
  - QR code size 1.5-2cm
  - with frame 2.25-3cm
  - Page size 18x27cm; 6x9 = 54
  """
  # this code currently does not work, but not used anyhow. Have to fix it in the future

  # import qrcode
  # import numpy as np
  # from PIL import Image
  # #from commonTools import commonTools as cT  # don't import globally since it disturbs translation
  # #img = qrcode.make(cT.uuidv4(), error_correction=qrcode.constants.ERROR_CORRECT_M)
  # size = -1 # img.size[0]
  # hSize = 6*size
  # vSize = 9*size
  # new_im = Image.new('RGB', (hSize, vSize))
  # for i in np.arange(0, hSize, size):
  #   for j in np.arange(0, vSize, size):
  #     #img = qrcode.make(cT.uuidv4(), error_correction=qrcode.constants.ERROR_CORRECT_M)
  #     # if j==0:  #make top row yellow
  #     #   data = np.array(img.convert("RGB"))
  #     #   red, green, blue = data.T
  #     #   mask = (red==255) & (green==255) & (blue==255)
  #     #   data[:,:,:][mask.T]=(255,255,0)
  #     #   img = Image.fromarray(data)
  #     # new_im.paste(img, (i, j))
  #     print(i,j)
  # new_im.save(fileName)
  return


def printQRcodeSticker(codes:list[list[str]]=[],
                       page:dict[str,Any]={'size':[991,306], 'tiles':2, 'margin': 60, 'font':40},
                       printer:dict[str,str]={'model':'QL-700', 'dev':'0x04f9:0x2042/3', 'size':'29x90'}) -> None:
  """
  Codes: key-value pairs of qr-code and label
   - filled to achieve tiles

  Sticker:
   - size: 90.3x29 mm => [991,306] px
   - tiles: number of items on the sticker
   - margin: margin between tiles
   - font: font size in px

  Printer:
   - model: brother label printer QL-700
   - dev: device in idVendor:idProduct/iSerial
     execute 'lsusb -v'; find printer
   - size: label size in mm
  """
  # this code currently does not work, but not used anyhow. Have to fix it in the future

  # import qrcode, tempfile
  # import numpy as np
  # from PIL import Image, ImageDraw, ImageFont
  # # fnt = ImageFont.truetype("arial.ttf", page['font'])
  # offset    = int((page['size'][0]+page['margin'])/page['tiles'])
  # qrCodeSize= min(offset-page['font']-page['margin'], page['size'][1])
  # print("Effective label size",page['size'], "offset",offset, 'qrCodeSize',qrCodeSize)
  # cropQRCode  = 40         #created by qrcode library
  # image = Image.new('RGBA', page['size'], color=(255,255,255,255) )
  # for numCodes, idx in enumerate(range(page['tiles'])):
  #   codeI, text = codes[idx] if idx<len(codes) else ('', '')
  #   print(text)
  #   if len(codeI)==0:
  #     codeI = uuid.uuid4().hex
  #   # add text: temporary comment out since library updated and functions changed
  #   # width, height = fnt.getsize(text)
  #   # txtImage = Image.new('L', (width, height), color=255)
  #   # d = ImageDraw.Draw(txtImage)
  #   # d.text( (0, 0), text,  font=fnt, fill=0)
  #   # txtImage=txtImage.rotate(90, expand=1)
  #   # if width>page['size'][1]:  #shorten it to fit into height
  #   #   txtImage=txtImage.crop((0,width-page['size'][1],height,width))
  #   # image.paste(txtImage, (numCodes*offset+qrCodeSize-4, int((page['size'][1]-txtImage.size[1])/2)   ))
  #   # add qrcode
  #   qrCode = qrcode.make(codeI, error_correction=qrcode.constants.ERROR_CORRECT_M)
  #   qrCode = qrCode.crop((cropQRCode, cropQRCode, qrCode.size[0]-cropQRCode, qrCode.size[0]-cropQRCode))
  #   qrCode = qrCode.resize((qrCodeSize, qrCodeSize))
  #   image.paste(qrCode, (numCodes*offset, int((page['size'][1]-qrCodeSize)/2)))
  # tmpFileName = tempfile.gettempdir()+os.sep+'tmpQRcode.png'
  # print('Create temp-file',tmpFileName)
  # image.save(tmpFileName)
  # cmd = 'brother_ql -b pyusb -m '+printer['model']+' -p usb://'+printer['dev']+' print -l '+printer['size']+' -r auto '+tmpFileName
  # reply = os.system(cmd)
  # if reply>0:
  #   print('**ERROR mpq01: Printing error')
  #   image.show()
  return



# Form implementation generated from reading ui file 'edit_metadata_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_EditMetadataDialog(object):
  def setupUi(self, EditMetadataDialog):
    EditMetadataDialog.setObjectName("EditMetadataDialog")
    EditMetadataDialog.resize(950, 428)
    self.verticalLayout = QtWidgets.QVBoxLayout(EditMetadataDialog)
    self.verticalLayout.setObjectName("verticalLayout")
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName("mainVerticalLayout")
    self.minimalFullComboBox = QtWidgets.QComboBox(parent=EditMetadataDialog)
    self.minimalFullComboBox.setObjectName("minimalFullComboBox")
    self.mainVerticalLayout.addWidget(self.minimalFullComboBox)
    self.horizontalLayout = QtWidgets.QHBoxLayout()
    self.horizontalLayout.setObjectName("horizontalLayout")
    self.label = QtWidgets.QLabel(parent=EditMetadataDialog)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
    self.label.setSizePolicy(sizePolicy)
    self.label.setMinimumSize(QtCore.QSize(100, 0))
    self.label.setObjectName("label")
    self.horizontalLayout.addWidget(self.label)
    self.licenseNameLineEdit = QtWidgets.QLineEdit(parent=EditMetadataDialog)
    self.licenseNameLineEdit.setClearButtonEnabled(True)
    self.licenseNameLineEdit.setObjectName("licenseNameLineEdit")
    self.horizontalLayout.addWidget(self.licenseNameLineEdit)
    self.licenseURLLineEdit = QtWidgets.QLineEdit(parent=EditMetadataDialog)
    self.licenseURLLineEdit.setClearButtonEnabled(True)
    self.licenseURLLineEdit.setObjectName("licenseURLLineEdit")
    self.horizontalLayout.addWidget(self.licenseURLLineEdit)
    self.mainVerticalLayout.addLayout(self.horizontalLayout)
    self.metadataBlockComboBox = QtWidgets.QComboBox(parent=EditMetadataDialog)
    self.metadataBlockComboBox.setObjectName("metadataBlockComboBox")
    self.mainVerticalLayout.addWidget(self.metadataBlockComboBox)
    self.typesComboBox = QtWidgets.QComboBox(parent=EditMetadataDialog)
    self.typesComboBox.setObjectName("typesComboBox")
    self.mainVerticalLayout.addWidget(self.typesComboBox)
    self.metadataScrollArea = QtWidgets.QScrollArea(parent=EditMetadataDialog)
    self.metadataScrollArea.setWidgetResizable(True)
    self.metadataScrollArea.setObjectName("metadataScrollArea")
    self.scrollAreaWidgetContents = QtWidgets.QWidget()
    self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 928, 249))
    self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
    self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
    self.verticalLayout_3.setObjectName("verticalLayout_3")
    self.metadataScrollVerticalLayout = QtWidgets.QVBoxLayout()
    self.metadataScrollVerticalLayout.setObjectName("metadataScrollVerticalLayout")
    self.verticalLayout_3.addLayout(self.metadataScrollVerticalLayout)
    self.metadataScrollArea.setWidget(self.scrollAreaWidgetContents)
    self.mainVerticalLayout.addWidget(self.metadataScrollArea)
    self.verticalLayout.addLayout(self.mainVerticalLayout)
    self.buttonBox = QtWidgets.QDialogButtonBox(parent=EditMetadataDialog)
    self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
    self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Save)
    self.buttonBox.setObjectName("buttonBox")
    self.verticalLayout.addWidget(self.buttonBox)

    self.retranslateUi(EditMetadataDialog)
    self.buttonBox.accepted.connect(EditMetadataDialog.accept) # type: ignore
    self.buttonBox.rejected.connect(EditMetadataDialog.reject) # type: ignore
    QtCore.QMetaObject.connectSlotsByName(EditMetadataDialog)

  def retranslateUi(self, EditMetadataDialog):
    _translate = QtCore.QCoreApplication.translate
    EditMetadataDialog.setWindowTitle(_translate("EditMetadataDialog", "Edit metadata"))
    EditMetadataDialog.setToolTip(_translate("EditMetadataDialog", "<html><head/><body><p><span style=\" font-style:italic;\">Edit the full/minimal set of metadata information for the datasets created in dataverse. Datasets corresponds to PASTA projects and needs to be mapped to the appropriate PASTA level properties as needed.</span></p></body></html>"))
    self.minimalFullComboBox.setToolTip(_translate("EditMetadataDialog", "Select either \"minimal\" or \"full\" from the given list. Below UI will be updated accordingly."))
    self.label.setText(_translate("EditMetadataDialog", "License"))
    self.licenseNameLineEdit.setToolTip(_translate("EditMetadataDialog", "Enter the license name. Default: CC0 1.0"))
    self.licenseNameLineEdit.setPlaceholderText(_translate("EditMetadataDialog", "Enter the license name."))
    self.licenseURLLineEdit.setToolTip(_translate("EditMetadataDialog", "Provide the license URL. Default: http://creativecommons.org/publicdomain/zero/1.0"))
    self.licenseURLLineEdit.setPlaceholderText(_translate("EditMetadataDialog", "Enter the license URL."))
    self.metadataBlockComboBox.setToolTip(_translate("EditMetadataDialog", "Select the metadata block to be edited."))
    self.typesComboBox.setToolTip(_translate("EditMetadataDialog", "Select the metadata block type to be edited."))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    EditMetadataDialog = QtWidgets.QDialog()
    ui = Ui_EditMetadataDialog()
    ui.setupUi(EditMetadataDialog)
    EditMetadataDialog.show()
    sys.exit(app.exec())



""" Represents the main dialog for dataverse integration. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: main_dialog.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import datetime
import logging
import textwrap
import time
from typing import Any

from PySide6 import QtWidgets
from PySide6.QtWidgets import QDialog, QFrame, QLabel, QMessageBox, QWidget

from pasta_eln.GUI.dataverse.completed_uploads import CompletedUploads
from pasta_eln.GUI.dataverse.config_dialog import ConfigDialog
from pasta_eln.GUI.dataverse.dialog_extension import DialogExtension
from pasta_eln.GUI.dataverse.edit_metadata_dialog import EditMetadataDialog
from pasta_eln.GUI.dataverse.main_dialog_base import Ui_MainDialogBase
from pasta_eln.GUI.dataverse.project_item_frame_base import Ui_ProjectItemFrame
from pasta_eln.GUI.dataverse.upload_config_dialog import UploadConfigDialog
from pasta_eln.GUI.dataverse.upload_widget_base import Ui_UploadWidgetFrame
from pasta_eln.backend import Backend
from pasta_eln.dataverse.config_model import ConfigModel
from pasta_eln.dataverse.data_upload_task import DataUploadTask
from pasta_eln.dataverse.database_api import DatabaseAPI
from pasta_eln.dataverse.project_model import ProjectModel
from pasta_eln.dataverse.task_thread_extension import TaskThreadExtension
from pasta_eln.dataverse.upload_model import UploadModel
from pasta_eln.dataverse.upload_queue_manager import UploadQueueManager
from pasta_eln.dataverse.upload_status_values import UploadStatusValues
from pasta_eln.dataverse.utils import check_if_dataverse_exists, check_if_minimal_metadata_exists, \
  check_login_credentials, get_formatted_message, update_status


class MainDialog(Ui_MainDialogBase):
  """
  Represents the main dialog.

  Explanation:
      This class represents the main dialog of the application.
      It provides methods to handle UI events and manage the upload process.

  """

  def __new__(cls, *_: Any, **__: Any) -> Any:
    """
    Creates a new instance of the MainDialog class.

    Explanation:
        This method creates a new instance of the MainDialog class.

    Args:
        *_: Variable length argument list.
        **__: Arbitrary keyword arguments.

    Returns:
        Any: The new instance of the MainDialog class.
    """
    return super(MainDialog, cls).__new__(cls)

  def __init__(self, backend: Backend) -> None:
    """
    Initializes the MainDialog.

    Explanation:
        This method initializes the MainDialog class and sets up the UI.
        It also initializes the database and populates the project widgets.

    """
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.instance = DialogExtension()
    super().setupUi(self.instance)
    self.backend = backend
    self.db_api = DatabaseAPI()
    self.db_api.initialize_database()
    self.is_dataverse_configured: tuple[bool, str] = self.check_if_dataverse_is_configured()
    self.config_dialog: ConfigDialog | None = None

    if self.is_dataverse_configured[0]:
      self.completed_uploads_dialog = CompletedUploads()
      self.edit_metadata_dialog = EditMetadataDialog()
      self.upload_manager_task = UploadQueueManager()
      self.config_upload_dialog = UploadConfigDialog(self.upload_manager_task.set_concurrent_uploads)
      self.upload_manager_task_thread = TaskThreadExtension(self.upload_manager_task)

      # Connect signals and slots
      self.uploadPushButton.clicked.connect(self.start_upload)
      self.clearFinishedPushButton.clicked.connect(self.clear_finished)
      self.selectAllPushButton.clicked.connect(lambda: self.select_deselect_all_projects(True))
      self.deselectAllPushButton.clicked.connect(lambda: self.select_deselect_all_projects(False))
      self.configureUploadPushButton.clicked.connect(self.show_configure_upload)
      self.showCompletedPushButton.clicked.connect(self.show_completed_uploads)
      self.editFullMetadataPushButton.clicked.connect(self.show_edit_metadata)
      self.cancelAllPushButton.clicked.connect(self.upload_manager_task.cancel_all_tasks.emit)
      self.buttonBox.button(QtWidgets.QDialogButtonBox.StandardButton.Cancel).clicked.connect(self.close_ui)
      self.instance.closed.connect(self.close_ui)

      # Load UI
      self.load_ui()

  def load_ui(self) -> None:
    """
    Loads the UI for the main dialog.

    Explanation:
        This method loads the UI for the main dialog by retrieving project models from the database
        and adding corresponding widgets to the projects scroll area.

    Args:
        self: The instance of the class.
    """
    for project in self.db_api.get_models(ProjectModel):
      if isinstance(project, ProjectModel):
        widget = self.get_project_widget(project)
        self.projectsScrollAreaVerticalLayout.addWidget(widget)

  def get_upload_widget(self, project_name: str = "") -> dict[str, QFrame | Ui_UploadWidgetFrame]:
    """
    Creates the upload widget for a project.

    Explanation:
        This method retrieves the upload widget for a specific project.
        It creates a QFrame instance and sets up the UI for the upload widget.

    Args:
        project_name (str): The name of the project.

    Returns:
        dict[str, QFrame | Ui_UploadWidgetFrame]: The upload widget.
    """
    base_frame = QtWidgets.QFrame()
    upload_widget_ui = Ui_UploadWidgetFrame()
    upload_widget_ui.setupUi(base_frame)
    upload_widget_ui.uploadProjectLabel.setText(textwrap.fill(project_name, 45, max_lines=1))
    upload_widget_ui.uploadProjectLabel.setToolTip(project_name)
    update_status(UploadStatusValues.Queued.name,
                  upload_widget_ui.statusLabel.setText,
                  upload_widget_ui.statusIconLabel.setPixmap)
    upload_widget_ui.logConsoleTextEdit.hide()
    upload_widget_ui.showLogPushButton.clicked.connect(lambda: self.show_hide_log(upload_widget_ui.showLogPushButton))
    upload_widget_ui.modelIdLabel.hide()
    return {"base": base_frame, "widget": upload_widget_ui}

  def get_project_widget(self, project: ProjectModel) -> QWidget:
    """
    Creates the project widget.


    Explanation:
        This method retrieves the project widget for a specific project.
        It creates a QWidget instance and sets up the UI for the project widget.

    Args:
        project (ProjectModel): The project model representing the project.

    Returns:
        QWidget: The project widget.

    Raises:
        ValueError: If the project model is not an instance of ProjectModel.
                    If the project model date is not in expected ISO format.
    """
    if not isinstance(project, ProjectModel):
      raise ValueError("Project model must be an instance of ProjectModel!")
    project_widget_frame = QtWidgets.QFrame()
    project_widget_ui = Ui_ProjectItemFrame()
    project_widget_ui.setupUi(project_widget_frame)
    project_widget_ui.projectNameLabel.setText(textwrap.fill(project.name or "", width=80, max_lines=1))
    project_widget_ui.projectNameLabel.setToolTip(project.name)
    project_widget_ui.modifiedDateTimeLabel.setText(
      datetime.datetime.fromisoformat(project.date or "").strftime("%Y-%m-%d %H:%M:%S"))
    project_widget_ui.projectDocIdLabel.hide()
    project_widget_ui.projectDocIdLabel.setText(project.id)
    return project_widget_frame

  def start_upload(self) -> None:
    """
    Starts the upload process.

    Explanation:
        This method starts the upload process for the selected projects.
        It retrieves the selected projects, creates upload widgets, and adds them to the upload queue.
        It also starts the upload manager task to process the upload queue.

    """
    if not self.check_if_minimal_metadata_present():
      self.logger.warning("Minimum metadata not present. Please add all needed metadata and then retry!")
      return
    for widget_pos in range(self.projectsScrollAreaVerticalLayout.count()):
      project_widget = self.projectsScrollAreaVerticalLayout.itemAt(widget_pos).widget()
      if project_widget.findChild(QtWidgets.QCheckBox, name="projectCheckBox").isChecked():
        upload_widget = self.get_upload_widget(
          project_widget.findChild(QtWidgets.QLabel, name="projectNameLabel").toolTip())
        self.uploadQueueVerticalLayout.addWidget(upload_widget["base"])
        widget = upload_widget["widget"]
        if not isinstance(widget, Ui_UploadWidgetFrame):
          continue
        project_id = project_widget.findChild(QtWidgets.QLabel, name="projectDocIdLabel").text()
        task_thread = TaskThreadExtension(
          DataUploadTask(widget.uploadProjectLabel.toolTip(),  # label text may be truncated,
                         #  but the tooltip contains the full project name
                         project_id,
                         widget.uploadProgressBar.setValue,
                         widget.statusLabel.setText,
                         widget.statusIconLabel.setPixmap,
                         widget.uploadCancelPushButton.clicked,
                         self.backend))
        self.upload_manager_task.add_to_queue(task_thread)
        if isinstance(task_thread.task, DataUploadTask):
          task_thread.task.upload_model_created.connect(widget.modelIdLabel.setText)
    self.upload_manager_task_thread.task.start.emit()

  def clear_finished(self) -> None:
    """
    Clears the finished upload widgets.
    Upload activity finishes in the following scenarios:
      Upload status is Finished, Error, Warning, or Canceled

    Explanation:
        This method clears the finished upload widgets from the upload queue.

    """
    for widget_pos in reversed(range(self.uploadQueueVerticalLayout.count())):
      project_widget = self.uploadQueueVerticalLayout.itemAt(widget_pos).widget()
      status_text = project_widget.findChild(QtWidgets.QLabel, name="statusLabel").text()
      if (status_text in [
        UploadStatusValues.Finished.name,
        UploadStatusValues.Error.name,
        UploadStatusValues.Warning.name,
        UploadStatusValues.Cancelled.name
      ]):
        project_widget.setParent(None)

  def select_deselect_all_projects(self, checked: bool) -> None:
    """
    Selects or deselects all projects.

    Explanation:
        This method selects or deselects all projects based on the provided checked value.
        It iterates over the project widgets and sets the checked state of the projectCheckBox.

    Args:
        checked (bool): The checked state to set for the projectCheckBox.

    """
    for widget_pos in range(self.projectsScrollAreaVerticalLayout.count()):
      project_widget = self.projectsScrollAreaVerticalLayout.itemAt(widget_pos).widget()
      project_widget.findChild(QtWidgets.QCheckBox, name="projectCheckBox").setChecked(checked)

  def show_configure_upload(self) -> None:
    """
    Shows the configure upload dialog.

    Explanation:
        This method shows the configure upload dialog.

    """
    self.config_upload_dialog.load_ui()
    self.config_upload_dialog.instance.show()

  def show_completed_uploads(self) -> None:
    """
    Shows the completed uploads dialog.

    Explanation:
        This method shows the completed uploads dialog.

    """
    self.completed_uploads_dialog.show()

  def show_edit_metadata(self) -> None:
    """
    Shows the edit metadata dialog.

    Explanation:
        This method shows the edit metadata dialog.

    """
    self.edit_metadata_dialog.show()

  def release_upload_manager(self) -> None:
    """
    Releases the upload manager.

    Explanation:
        This method releases the upload manager by quitting the upload manager task thread.

    """
    self.upload_manager_task_thread.quit()

  def close_ui(self) -> None:
    """
    Closes the UI.

    Explanation:
        This method closes the UI by quitting the upload manager task thread and adding a delay of 0.5 seconds.

    """
    self.upload_manager_task_thread.quit()
    time.sleep(0.5)

  def show_hide_log(self, button: QtWidgets.QPushButton) -> None:
    """
    Shows or hides the log console.

    Explanation:
        This method shows or hides the log console based on the current state.
        It retrieves the log console text edit widget and toggles its visibility.
        It also retrieves the model ID label and updates the log console text based on the model ID.

    Args:
        button (QtWidgets.QPushButton): The button triggering the action.

    """
    frame = button.parent()
    log_console_text_edit = frame.findChild(QtWidgets.QTextEdit, name="logConsoleTextEdit")
    if isinstance(log_console_text_edit, QtWidgets.QTextEdit):
      if log_console_text_edit.isHidden():
        log_console_text_edit.show()
      else:
        log_console_text_edit.hide()
      id_label = frame.findChild(QtWidgets.QLabel, name="modelIdLabel")
      if upload_model_id := id_label.text() if isinstance(id_label, QtWidgets.QLabel) else "":
        model = self.db_api.get_model(upload_model_id, UploadModel)
        log_console_text_edit.setText(model.log if isinstance(model, UploadModel) else "")

  def check_if_minimal_metadata_present(self) -> bool:
    """
    Checks if minimal metadata is present.

    Returns:
        bool: True if minimal metadata is present, False otherwise.

    Explanation:
        This function checks if minimal metadata is present in the configuration model.
        It returns True if minimal metadata is present, and False otherwise.
        If minimal metadata is missing, it displays a warning message to the user.
    """
    self.logger.info("Checking if minimal metadata is present...")
    config_model = self.db_api.get_config_model()
    metadata_exists = False
    if config_model is None:
      self.logger.error("Failed to load config model!")
      return metadata_exists
    if isinstance(config_model, ConfigModel) and isinstance(config_model.metadata, dict):
      if missing_metadata := check_if_minimal_metadata_exists(self.logger,
                                                              config_model.metadata,
                                                              False):
        message = get_formatted_message(missing_metadata)
        metadata_exists = message == ""
        if not metadata_exists:
          self.show_message(self.instance, "Missing Minimal Metadata", message)
    return metadata_exists

  def check_if_dataverse_is_configured(self) -> tuple[bool, str]:
    """
    Checks if dataverse is configured.

    Explanation:
        This function checks if the dataverse is configured by retrieving configuration details from the database.
        It validates the configuration, including the server URL, API token, and dataverse ID, and returns a success status.

    Returns:
        tuple[bool, str]: A tuple containing a boolean indicating success and a message string.
    """
    self.logger.info("Checking if dataverse is configured..")
    config_model = self.db_api.get_config_model()
    if config_model is None or config_model.dataverse_login_info is None:
      self.logger.error("Failed to load config model!")
      return False, "Failed to load config model!"
    server_url = config_model.dataverse_login_info["server_url"]
    api_token = config_model.dataverse_login_info["api_token"]
    dataverse_id = config_model.dataverse_login_info["dataverse_id"]
    success, _ = check_login_credentials(self.logger, api_token, server_url)
    if not success:
      return False, "Please re-enter the correct API token / server URL via the configuration dialog."
    if success := success and check_if_dataverse_exists(
        self.logger, api_token, server_url, dataverse_id):
      return success, "Dataverse configured successfully!"
    else:
      return False, f"Please re-enter the correct dataverse ID via the configuration dialog, Saved id: {dataverse_id} is incorrect!"

  def show(self) -> None:
    """
    Shows the instance.

    Explanation:
        This method shows the instance by calling its show method if the dataverse is configured.
        If the dataverse is not configured, it displays a message and shows the ConfigDialog.

    Args:
        self: The instance of the class.
    """
    self.logger.info("Show MainDialog UI..")
    if self.is_dataverse_configured[0]:
      self.instance.show()
    else:
      self.config_dialog = ConfigDialog()
      self.config_dialog.show()

  def show_message(self,
                   parent: QDialog,
                   title: str,
                   message: str,
                   icon: QMessageBox.Icon = QMessageBox.Icon.Warning) -> None:
    """
    Shows a message box with the specified title, message, and icon.

    Explanation:
        This function creates a message box with the provided title, message, and icon, and displays it to the user.
        It also adjusts the width of the message box label to fit the content properly.

    Args:
        self: The instance of the class.
        parent (QDialog): The parent dialog for the message box.
        title (str): The title of the message box.
        message (str): The message content to display.
        icon (QMessageBox.Icon, optional): The icon to display in the message box (default is QMessageBox.Warning).
    """

    msg_box = QMessageBox(parent)
    msg_box.setWindowTitle(title)
    msg_box.setIcon(icon or QMessageBox.Icon.Warning)
    msg_box.setText(message)
    qt_msgbox_label: QLabel | object = msg_box.findChild(QLabel, "qt_msgbox_label")
    if not isinstance(qt_msgbox_label, QLabel):
      self.logger.error("Failed to find message box label!")
      return
    width = qt_msgbox_label.fontMetrics().boundingRect(qt_msgbox_label.text()).width()
    qt_msgbox_label.setFixedWidth(width)
    msg_box.exec()



""" LookupIriAction class which is extended from the QAction class and contains IRI lookup logic"""
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: lookup_iri_action.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from typing import Union

from PySide6.QtCore import QModelIndex, QPersistentModelIndex, Qt
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QLineEdit, QWidget
from qtawesome import icon

from pasta_eln.GUI.data_hierarchy.terminology_lookup_dialog import TerminologyLookupDialog


class LookupIriAction(QAction):
  """ LookupIriAction class which is extended from the QAction class and contains IRI lookup logic"""

  def __init__(self,
               lookup_term: str | None = None,
               parent_line_edit: QLineEdit | None = None,
               cell_index: Union[QModelIndex, QPersistentModelIndex, None] = None) -> None:
    """
    Initializes the LookupIriAction
    Args:
      lookup_term (object): Search term used for the lookup
      parent_line_edit (QWidget): Parent line edit whose text is to be updated when lookup results are available
      cell_index (Union[QModelIndex, QPersistentModelIndex]): Cell index of the table view whose value is to be updated when lookup results are available
    """
    super().__init__(
      icon=icon('fa.angle-right', color='black', scale_factor=1),
      text="Lookup IRI online",
      parent=parent_line_edit)
    self.parent_line_edit: QLineEdit | None = parent_line_edit
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.terminology_lookup_dialog = TerminologyLookupDialog(lookup_term, self.terminology_lookup_accepted_callback)
    self.cell_index: Union[QModelIndex, QPersistentModelIndex, None] = cell_index
    super().triggered.connect(self.show_terminology_lookup_dialog)

  def show_terminology_lookup_dialog(self) -> None:
    """
    Retrieves the IRI using terminology lookup widget
    Returns: Nothing

    """
    self.logger.info("Lookup dialog shown..")
    self.terminology_lookup_dialog.show()

  def terminology_lookup_accepted_callback(self) -> None:
    """
    Callback for the terminology lookup accepted button
    Returns: Nothing
    """
    self.logger.info("Accepted IRIs: %s", self.terminology_lookup_dialog.selected_iris)
    iris = " ".join(self.terminology_lookup_dialog.selected_iris)
    if self.parent_line_edit:
      self.parent_line_edit.setText(iris)
    if self.cell_index and self.cell_index.isValid():
      self.cell_index.model().setData(self.cell_index, iris, Qt.ItemDataRole.UserRole)



# Form implementation generated from reading ui file 'upload_widget_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_UploadWidgetFrame(object):
  def setupUi(self, UploadWidgetFrame):
    UploadWidgetFrame.setObjectName("UploadWidgetFrame")
    UploadWidgetFrame.resize(1141, 228)
    UploadWidgetFrame.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
    UploadWidgetFrame.setFrameShadow(QtWidgets.QFrame.Shadow.Raised)
    self.gridLayout = QtWidgets.QGridLayout(UploadWidgetFrame)
    self.gridLayout.setObjectName("gridLayout")
    self.uploadWidgetMainverticalLayout = QtWidgets.QVBoxLayout()
    self.uploadWidgetMainverticalLayout.setObjectName("uploadWidgetMainverticalLayout")
    self.uploadWidgetHorizontalLayout = QtWidgets.QHBoxLayout()
    self.uploadWidgetHorizontalLayout.setObjectName("uploadWidgetHorizontalLayout")
    self.uploadProjectLabel = QtWidgets.QLabel(parent=UploadWidgetFrame)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.uploadProjectLabel.sizePolicy().hasHeightForWidth())
    self.uploadProjectLabel.setSizePolicy(sizePolicy)
    self.uploadProjectLabel.setMinimumSize(QtCore.QSize(300, 0))
    self.uploadProjectLabel.setMaximumSize(QtCore.QSize(300, 100))
    self.uploadProjectLabel.setObjectName("uploadProjectLabel")
    self.uploadWidgetHorizontalLayout.addWidget(self.uploadProjectLabel)
    self.uploadProgressBar = QtWidgets.QProgressBar(parent=UploadWidgetFrame)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.uploadProgressBar.sizePolicy().hasHeightForWidth())
    self.uploadProgressBar.setSizePolicy(sizePolicy)
    self.uploadProgressBar.setMinimumSize(QtCore.QSize(400, 0))
    self.uploadProgressBar.setProperty("value", 0)
    self.uploadProgressBar.setObjectName("uploadProgressBar")
    self.uploadWidgetHorizontalLayout.addWidget(self.uploadProgressBar)
    self.statusIconLabel = QtWidgets.QLabel(parent=UploadWidgetFrame)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
    sizePolicy.setHorizontalStretch(50)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.statusIconLabel.sizePolicy().hasHeightForWidth())
    self.statusIconLabel.setSizePolicy(sizePolicy)
    self.statusIconLabel.setMinimumSize(QtCore.QSize(25, 25))
    self.statusIconLabel.setText("")
    self.statusIconLabel.setObjectName("statusIconLabel")
    self.uploadWidgetHorizontalLayout.addWidget(self.statusIconLabel)
    self.statusLabel = QtWidgets.QLabel(parent=UploadWidgetFrame)
    self.statusLabel.setMinimumSize(QtCore.QSize(0, 0))
    self.statusLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeading|QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignVCenter)
    self.statusLabel.setObjectName("statusLabel")
    self.uploadWidgetHorizontalLayout.addWidget(self.statusLabel)
    self.showLogPushButton = QtWidgets.QPushButton(parent=UploadWidgetFrame)
    self.showLogPushButton.setMinimumSize(QtCore.QSize(100, 0))
    icon = QtGui.QIcon.fromTheme("document-page-setup")
    self.showLogPushButton.setIcon(icon)
    self.showLogPushButton.setObjectName("showLogPushButton")
    self.uploadWidgetHorizontalLayout.addWidget(self.showLogPushButton)
    self.uploadCancelPushButton = QtWidgets.QPushButton(parent=UploadWidgetFrame)
    self.uploadCancelPushButton.setMinimumSize(QtCore.QSize(100, 0))
    icon = QtGui.QIcon.fromTheme("window-close")
    self.uploadCancelPushButton.setIcon(icon)
    self.uploadCancelPushButton.setObjectName("uploadCancelPushButton")
    self.uploadWidgetHorizontalLayout.addWidget(self.uploadCancelPushButton)
    self.uploadWidgetMainverticalLayout.addLayout(self.uploadWidgetHorizontalLayout)
    self.logConsoleTextEdit = QtWidgets.QTextEdit(parent=UploadWidgetFrame)
    self.logConsoleTextEdit.setEnabled(True)
    self.logConsoleTextEdit.setMinimumSize(QtCore.QSize(0, 150))
    self.logConsoleTextEdit.setReadOnly(True)
    self.logConsoleTextEdit.setObjectName("logConsoleTextEdit")
    self.uploadWidgetMainverticalLayout.addWidget(self.logConsoleTextEdit)
    self.modelIdLabel = QtWidgets.QLabel(parent=UploadWidgetFrame)
    self.modelIdLabel.setText("")
    self.modelIdLabel.setObjectName("modelIdLabel")
    self.uploadWidgetMainverticalLayout.addWidget(self.modelIdLabel)
    self.gridLayout.addLayout(self.uploadWidgetMainverticalLayout, 0, 0, 1, 1)

    self.retranslateUi(UploadWidgetFrame)
    QtCore.QMetaObject.connectSlotsByName(UploadWidgetFrame)

  def retranslateUi(self, UploadWidgetFrame):
    _translate = QtCore.QCoreApplication.translate
    UploadWidgetFrame.setWindowTitle(_translate("UploadWidgetFrame", "Frame"))
    self.uploadProjectLabel.setToolTip(_translate("UploadWidgetFrame", "Name of the PASTA project enqueued for dataverse upload."))
    self.uploadProjectLabel.setText(_translate("UploadWidgetFrame", "Example Project 1"))
    self.uploadProgressBar.setToolTip(_translate("UploadWidgetFrame", "Shows the upload progress."))
    self.statusLabel.setToolTip(_translate("UploadWidgetFrame", "Displays the status of the upload process (Queued/Success/Failed/In-progress)."))
    self.statusLabel.setText(_translate("UploadWidgetFrame", "Queued..."))
    self.showLogPushButton.setToolTip(_translate("UploadWidgetFrame", "Click to view the log for this particular upload."))
    self.showLogPushButton.setText(_translate("UploadWidgetFrame", "Show Log"))
    self.uploadCancelPushButton.setToolTip(_translate("UploadWidgetFrame", "Click to cancel this particular upload."))
    self.uploadCancelPushButton.setText(_translate("UploadWidgetFrame", "Cancel"))
    self.logConsoleTextEdit.setToolTip(_translate("UploadWidgetFrame", "Displays the generated log messages for this particular upload."))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    UploadWidgetFrame = QtWidgets.QFrame()
    ui = Ui_UploadWidgetFrame()
    ui.setupUi(UploadWidgetFrame)
    UploadWidgetFrame.show()
    sys.exit(app.exec())



""" Class for simulating progress updates in the background. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: progress_updater_thread.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import random
import time

from PySide6 import QtCore
from PySide6.QtCore import QThread


class ProgressUpdaterThread(QThread):
  """
   Class for simulating progress updates in the background.

   Explanation:
       This class represents a thread for simulating progress updates in the background.
       It inherits from the QThread class and provides methods for simulating progress updates and emitting signals.
   """
  progress_update = QtCore.Signal(int)
  cancel = QtCore.Signal()
  finalize = QtCore.Signal()
  end = QtCore.Signal()

  def __init__(self) -> None:
    """
    Initialize the progress updater thread.

    Args:
        self: The instance of the progress updater thread.
    """
    super().__init__()
    self.cancelled = False
    self.finalized = False
    self.cancel.connect(self.cancel_progress)
    self.finalize.connect(self.finalize_progress)

  def run(self) -> None:
    """
    Simulates progress updates and emits progress signals until completion or interruption.

    Explanation:
        This method simulates progress updates in the background, emitting signals for each progress step.
        It emits progress updates until the task is cancelled or finalized.
        It waits for a total of 120 seconds before emitting the final progress signal and exiting.
    """
    # Simulate progress updates in the background for around 20 seconds
    for progress in range(1, 98, random.randint(1, 5)):
      if self.cancelled or self.finalized:
        break
      self.progress_update.emit(progress)
      time.sleep(1)

    # Wait for 100 seconds more before emitting the final progress
    for _ in range(1, 100):
      if self.cancelled or self.finalized:
        break
      time.sleep(1)
    if not self.cancelled:
      self.progress_update.emit(100)
    self.end.emit()

  def cancel_progress(self) -> None:
    """
    Sets the flag to indicate that the progress should be cancelled.

    Explanation:
        This method sets the 'finish' flag to True, indicating that the progress should be terminated.
    """
    self.cancelled = True

  def finalize_progress(self) -> None:
    """
    Finalizes the progress update.

    Explanation:
        This function sets the 'finalized' attribute to True, indicating that the progress update is complete.
    """
    self.finalized = True



""" Table Header dialog: change which columns are shown and in which order """
import json
import platform
from enum import Enum
from pathlib import Path
from typing import Any

import qrcode
from PIL.ImageQt import ImageQt
from PySide6.QtGui import QPixmap, QRegularExpressionValidator  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import QComboBox, QDialog, QDialogButtonBox, QFileDialog, QFormLayout, QGroupBox, QLabel, \
  QLineEdit, QMessageBox, QVBoxLayout  # pylint: disable=no-name-in-module
from cloudant.client import CouchDB

from ..guiCommunicate import Communicate
from ..guiStyle import IconButton, Label, TextButton, showMessage, widgetAndLayout
from ..miscTools import restart, upIn, upOut
from ..serverActions import passwordDecrypt, testLocal, testRemote


class ProjectGroup(QDialog):
  """ Table Header dialog: change which columns are shown and in which order """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      backend (Backend): PASTA-ELN backend
    """
    super().__init__()
    self.comm    = comm
    self.configuration = self.comm.backend.configuration

    # GUI elements
    self.setWindowTitle('Define and use project groups')
    self.setMinimumWidth(1000)
    mainL = QVBoxLayout(self)
    Label('Project group', 'h1', mainL)
    _, topbarL = widgetAndLayout('H', mainL, spacing='m')
    self.selectGroup = QComboBox()
    self.selectGroup.addItems(self.configuration['projectGroups'].keys())
    self.selectGroup.currentTextChanged.connect(self.changeProjectGroup)
    topbarL.addWidget(self.selectGroup)
    TextButton('New',         self, [Command.NEW],       topbarL)
    TextButton('Fill remote', self, [Command.FILL],      topbarL)
    TextButton('Create QR',   self, [Command.CREATE_QR], topbarL)
    TextButton('Check All',   self, [Command.CHECK],     topbarL)
    self.projectGroupName = QLineEdit('')
    self.projectGroupName.hide()
    mainL.addWidget(self.projectGroupName)
    _, bodyL = widgetAndLayout('H', mainL)
    #local
    localW = QGroupBox('Local credentials')
    localL = QFormLayout(localW)
    self.userNameL = QLineEdit('')
    self.userNameL.setValidator(QRegularExpressionValidator("[\\w.]{5,}"))
    localL.addRow('User name', self.userNameL)
    self.passwordL = QLineEdit('')
    self.passwordL.setValidator(QRegularExpressionValidator("\\S{5,}"))
    self.passwordL.setEchoMode(QLineEdit.EchoMode.PasswordEchoOnEdit)
    localL.addRow('Password', self.passwordL)
    self.databaseL = QLineEdit('')
    self.databaseL.setValidator(QRegularExpressionValidator("\\w{5,}"))
    localL.addRow('Database', self.databaseL)
    pathW, pathL = widgetAndLayout('H', spacing='s')
    self.pathL = QLineEdit('')
    pathL.addWidget(self.pathL, stretch=5)
    if platform.system()=='Windows':
      self.pathL.setValidator(QRegularExpressionValidator(r"[\\/~][\\w\\\\\\/:\.~]{5,}"))
    else:
      self.pathL.setValidator(QRegularExpressionValidator(r"[\\/~][\\w\\/]{5,}"))
    IconButton('fa5.folder-open', self, [Command.OPEN_DIR], pathL, 'Folder to save data in')
    localL.addRow('Path', pathW)
    bodyL.addWidget(localW)
    #remote
    remoteW = QGroupBox('Remote credentials')
    remoteL = QFormLayout(remoteW)
    self.userNameR = QLineEdit('')
    self.userNameR.setValidator(QRegularExpressionValidator("[\\w.]{5,}"))
    remoteL.addRow('User name', self.userNameR)
    self.passwordR = QLineEdit('')
    self.passwordR.setValidator(QRegularExpressionValidator("\\S{5,}"))
    self.passwordR.setEchoMode(QLineEdit.EchoMode.PasswordEchoOnEdit)
    remoteL.addRow('Password', self.passwordR)
    self.databaseR = QLineEdit('')
    self.databaseR.setValidator(QRegularExpressionValidator("\\w{5,}"))
    remoteL.addRow('Database', self.databaseR)
    self.serverR = QLineEdit('')
    self.serverR.setValidator(QRegularExpressionValidator("http:\\/\\/(?:[0-9]{1,3}\\.){3}[0-9]{1,3}:5984"))
    remoteL.addRow('Server', self.serverR)
    bodyL.addWidget(remoteW)
    #image
    self.image = QLabel()
    bodyL.addWidget(self.image)

    #final button box
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.addButton('Save encrypted', QDialogButtonBox.ButtonRole.AcceptRole)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)
    self.selectGroup.currentTextChanged.emit(self.configuration['defaultProjectGroup']) #emit to fill initially


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
    elif 'Save' in btn.text() and self.checkEntries():
      name = self.projectGroupName.text() if self.selectGroup.isHidden() else self.selectGroup.currentText()
      local = remote = {}
      if btn.text().endswith('Save'):
        localPath = self.pathL.text()
        if localPath.startswith('~'):
          localPath = (Path.home()/localPath[1:]).as_posix()
        local = {'user':self.userNameL.text(), 'password':self.passwordL.text(), \
                  'database':self.databaseL.text(), 'path':localPath}
        remote = {'user':self.userNameR.text(), 'password':self.passwordR.text(), \
                  'database':self.databaseR.text(), 'url':self.serverR.text()}
      elif btn.text().endswith('Save encrypted'):
        credL = upIn(f'{self.userNameL.text()}:{self.passwordL.text()}')
        credR = upIn(f'{self.userNameR.text()}:{self.passwordR.text()}')
        local = {'cred':credL, 'database':self.databaseL.text(), 'path':self.pathL.text()}
        remote = {'cred':credR, 'database':self.databaseR.text(), 'url':self.serverR.text()}
      newGroup = {'local':local, 'remote':remote}
      self.configuration['projectGroups'][name] = newGroup
      self.configuration['defaultProjectGroup'] = name
      with open(Path.home()/'.pastaELN.json', 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.configuration,indent=2))
      restart()
    else:
      print('dialogProjectGroup: did not get a fitting btn ',btn.text())
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.NEW:
      self.selectGroup.hide()
      self.projectGroupName.show()
      self.projectGroupName.setText('my_project_group_name')
      defaultProjectGroup = self.configuration['defaultProjectGroup']
      config = self.configuration['projectGroups'][defaultProjectGroup]
      if 'cred' in config['local']:
        u,p = upOut(config['local']['cred'])[0].split(':')
      else:
        u,p = config['local']['user'], config['local']['password']
      self.userNameL.setText(u)
      self.userNameR.setText('')
      self.passwordL.setText(p)
      self.passwordR.setText('')
      self.databaseL.setText('')
      self.databaseR.setText('')
      self.pathL.setText('')
      self.serverR.setText('')
    elif command[0] is Command.FILL:
      contentFile = QFileDialog.getOpenFileName(self, "Load remote credentials", str(Path.home()), '*.key')[0]
      with open(contentFile, encoding='utf-8') as fIn:
        content = json.loads( passwordDecrypt(bytes(fIn.read(), 'UTF-8')) )
        self.userNameR.setText(content['user-name'])
        self.passwordR.setText(content['password'])
        self.databaseR.setText(content['database'])
        self.serverR.setText(content['Server'])
    elif command[0] is Command.CREATE_QR:
      if self.projectGroupName.isHidden():
        configname = self.selectGroup.currentText()
      else:
        configname = self.projectGroupName.text()
      qrCode = {"configname": configname, "credentials":{"server":self.serverR.text(), \
          "username":self.userNameR.text(), "password":self.passwordR.text(), "database":self.databaseR.text()}}
      img = qrcode.make(json.dumps(qrCode), error_correction=qrcode.constants.ERROR_CORRECT_M)
      pixmap = QPixmap.fromImage(ImageQt(img).scaledToWidth(200))
      self.image.setPixmap(pixmap)
    elif command[0] is Command.CHECK:
      self.checkEntries()
    elif command[0] is Command.OPEN_DIR:
      if dirName := QFileDialog.getExistingDirectory(self, 'Choose directory to save data', str(Path.home())):
        self.pathL.setText(dirName)
    else:
      print("**ERROR projectGroup unknown:",command)
    return


  def checkEntries(self) -> bool:
    """
    Check if entries are ok

    Returns:
      bool: success
    """
    # local
    localTest = testLocal(self.userNameL.text(), self.passwordL.text(), self.databaseL.text())
    if 'Error: Local database does not exist' in localTest and \
       'success: Local username and password ok' in localTest and self.databaseL.text():
      button = QMessageBox.question(self, "Question", "Local database does not exist. Should I create it?")
      if button == QMessageBox.StandardButton.Yes:
        localTest += '  Local data was created\n'
        client = CouchDB(self.userNameL.text(), self.passwordL.text(), url='http://127.0.0.1:5984', connect=True)
        client.create_database(self.databaseL.text())
      else:
        localTest += '  Local data was NOT created\n'
    if not self.pathL.text():
      localTest += 'ERROR: Local path not given\n'
    else:
      fullLocalPath = self.comm.backend.basePath/self.pathL.text()
      if fullLocalPath.is_dir():
        localTest += 'success: Local path exists\n'
      else:
        button = QMessageBox.question(self, "Question", "Local folder does not exist. Should I create it?")
        if button == QMessageBox.StandardButton.Yes:
          fullLocalPath.mkdir()
        else:
          localTest += 'ERROR: Local path does not exist\n'
    # remote
    remoteTest = ''
    if self.userNameR.text()!='' and self.passwordR.text()!='' and self.databaseR.text()!='' and \
        self.serverR.text()!='':
      remoteTest = testRemote(self.serverR.text(), self.userNameR.text(), self.passwordR.text(), \
        self.databaseR.text())
    # give output
    if 'ERROR' in localTest or 'ERROR' in remoteTest:
      showMessage(self, 'ERROR occurred', localTest+remoteTest, 'Critical')
      return False
    #success
    showMessage(self, 'Successful test', localTest+remoteTest, 'Information')
    return True


  def changeProjectGroup(self, item:str) -> None:
    """
    change the project group to this; do not save to file

    Args:
      item (str): name of project group
    """
    config = self.configuration['projectGroups'][item]
    if 'cred' in config['local']:
      u,p = upOut(config['local']['cred'])[0].split(':')
    else:
      u,p = config['local']['user'], config['local']['password']
    self.userNameL.setText(u)
    self.passwordL.setText(p)
    self.databaseL.setText(config['local']['database'])
    self.pathL.setText(config['local']['path'])
    if 'url' in config['remote']:
      if 'cred' in config['remote']:
        u,p = upOut(config['remote']['cred'])[0].split(':')
      else:
        u,p = config['remote']['user'], config['remote']['password']
      self.userNameR.setText(u)
      self.passwordR.setText(p)
      self.databaseR.setText(config['remote']['database'])
      self.serverR.setText(config['remote']['url'])
    return


class Command(Enum):
  """ Commands used in this file """
  NEW       = 1
  FILL      = 2
  CREATE_QR = 3
  CHECK     = 4
  OPEN_DIR  = 5



""" Provides an interface to interact with a database. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: base_database_api.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import logging
from threading import Lock
from typing import Any

from cloudant import couchdb
from cloudant.database import CloudantDatabase
from cloudant.design_document import DesignDocument
from cloudant.document import Document
from cloudant.error import CloudantClientException, CloudantDatabaseException
from cloudant.query import Query
from cloudant.result import Result
from cloudant.view import View

from pasta_eln.dataverse.database_error import DatabaseError
from pasta_eln.dataverse.incorrect_parameter_error import IncorrectParameterError
from pasta_eln.dataverse.utils import log_and_create_error


class BaseDatabaseAPI:
  """
  Provides an interface to interact with a database.

  Explanation:
      This class encapsulates the functionality to interact with a database, including creating, retrieving, and updating documents,
      adding views to design documents, and getting view results.

  """

  def __init__(self,
               hostname: str,
               port: int,
               username: str,
               password: str) -> None:
    """
    Explanation:
        This method initializes the instance with the provided attributes for database connection.
        It sets up the logger, host, port, URL, password, username, and update lock for the database connection.

    Args:
        hostname (str): The hostname for the database connection.
        port (int): The port number for the database connection.
        username (str): The username for the database connection.
        password (str): The password for the database connection.

    Exceptions:
        IncorrectParameterError: If any of the parameters are of the incorrect type.
    """

    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    if isinstance(hostname, str):
      self.host: str = hostname
    else:
      raise IncorrectParameterError("Hostname must be a string")
    if isinstance(port, int):
      self.port: int = port
    else:
      raise IncorrectParameterError("Port must be an integer")
    self.url = f'http://{self.host}:{self.port}'
    if isinstance(password, str):
      self.password = password
    else:
      raise IncorrectParameterError("Password must be a string")
    if isinstance(username, str):
      self.username = username
    else:
      raise IncorrectParameterError("Username must be a string")
    self.update_lock = Lock()

  def create_database(self, db_name: str) -> CloudantDatabase:
    """
    Creates a database.

    Explanation:
        This function creates a database with the provided name using the Cloudant client.
        It logs the database creation process and handles exceptions if the database already exists.

    Args:
        db_name (str): The name of the database to be created.

    Returns:
        CloudantDatabase: The created database instance.
    """
    self.logger.info("Creating database with name : %s", db_name)
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      try:
        db = client.create_database(db_name, throw_on_exists=True)
      except CloudantClientException as e:
        self.logger.warning("Error creating database: %s", e)
        db = client[db_name]  # DB exists already
      return db

  def create_document(self, data: dict[str, Any], db_name: str) -> Document:
    """
    Creates a document in the database with the provided data.

    Args:
        db_name (str): The name of the database.
        data (dict[str, Any]): The data to be stored in the document.

    Returns:
        Document: The created document.

    """
    self.logger.info("Creating document with data: %s in database: %s",
                     data,
                     db_name)
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      document = None
      try:
        document = pasta_db.create_document(data, throw_on_exists=True)
      except CloudantDatabaseException as e:
        self.logger.error("Error creating document: %s", e)
      return document

  def get_document(self, document_id: str, db_name: str) -> Document:
    """
    Retrieves a document from the database based on the provided document ID.

    Args:
        db_name (str): The name of the database.
        document_id (str): The ID of the document to retrieve.

    Returns:
        Document: The retrieved document.

    Raises:
        DatabaseError: If the document ID is empty or if an error occurs during retrieval.

    """
    self.logger.info("Retrieving document with id: %s from database: %s", document_id, db_name)
    if not document_id:
      raise log_and_create_error(self.logger, DatabaseError, "Document ID cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      document = None
      try:
        document = pasta_db[document_id]
      except KeyError as e:
        self.logger.error("Error retrieving document: %s", e)
      return document

  def update_document(self, data: dict[str, Any], db_name: str) -> None:
    """
    Updates a document in the database with the provided data.
    Make sure to use the `_id` and `_rev` keys in the data to identify the document to update otherwise a new document will be created.

    Args:
        db_name (str): The name of the database.
        data (dict[str, Any]): The data to update the document with.
    """
    self.logger.info("Updating document with data: %s in database: %s", data, db_name)
    with self.update_lock:
      with couchdb(self.username,
                   self.password,
                   url=self.url,
                   connect=True) as client:
        pasta_db = client[db_name]
        with Document(pasta_db, data['_id']) as document:
          for key, value in data.items():
            if key not in ['_id', '_rev']:
              document[key] = value

  def add_view(self,
               db_name: str,
               design_document_name: str,
               view_name: str,
               map_func: str,
               reduce_func: str | None = None) -> None:
    """
    Adds a view to the specified design document in the database.

    Args:
        db_name (str): The name of the database.
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.
        map_func (str): The map function for the view.
        reduce_func (str, optional): The reduce function for the view. Defaults to None.

    Raises:
        ValueError: If design_document_name, view_name, or map_func is None.

    Examples:
        # Add a view with a map function
        add_view("my_design_doc", "my_view", "function(doc) { emit(doc._id, 1); }")

        # Add a view with a map and reduce function
        add_view("my_design_doc", "my_view", "function(doc) { emit(doc.name, doc.age); }", "_count")
    """
    self.logger.info("Adding view: %s to design document: %s, map_func: %s, reduce_func: %s in database: %s",
                     view_name,
                     design_document_name,
                     map_func,
                     reduce_func,
                     db_name)
    if design_document_name is None or view_name is None or map_func is None:
      raise log_and_create_error(self.logger, DatabaseError,
                                 "Design document name, view name, and map function cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        design_doc.add_view(view_name, map_func, reduce_func)

  def get_view(self,
               db_name: str,
               design_document_name: str,
               view_name: str) -> View:
    """
    Retrieves a view from the design document.

    Explanation:
        This method retrieves a view from the specified design document.
        It connects to the CouchDB instance, selects the appropriate database,
        and retrieves the view from the design document.

    Args:
        db_name (str): The name of the database.
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.

    Raises:
        ValueError: If the design document name or view name is empty.

    Returns:
        View: The retrieved view.
    """
    self.logger.info("Retrieving view: %s from design document: %s in database: %s",
                     view_name,
                     design_document_name,
                     db_name)
    if design_document_name is None or view_name is None:
      raise log_and_create_error(self.logger, DatabaseError, "Design document name, view name cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        return design_doc.get_view(view_name)

  def get_view_results(self,
                       db_name: str,
                       design_document_name: str,
                       view_name: str,
                       map_func: str | None = None,
                       reduce_func: str | None = None) -> list[dict[str, Any]]:
    """
    Gets the results of a view from the specified design document in the database.

    Args:
        db_name (str): The name of the database.
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.
        map_func (str, optional): The map function for the view. Defaults to None.
        reduce_func (str, optional): The reduce function for the view. Defaults to None.

    Raises:
        ValueError: If design_document_name or view_name is None.

    Returns:
        list[dict[str, Any]]: The results of the view as a list of dictionaries.

    Examples:
        # Get results from a view without map and reduce functions
        results = get_view_results("my_design_doc", "my_view")

        # Get results from a view with map and reduce functions
        results = get_view_results("my_design_doc", "my_view", "function(doc) { emit(doc._id, 1); }", "_count")
    """
    self.logger.info("Getting view results: %s from design document: %s, map_func: %s, reduce_func: %s in database: %s",
                     view_name,
                     design_document_name,
                     map_func,
                     reduce_func,
                     db_name)
    if design_document_name is None or view_name is None:
      raise log_and_create_error(self.logger, DatabaseError, "Design document name and view name cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      results: list[dict[str, Any]] = []
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        view = View(design_doc, view_name, map_func, reduce_func)
        results.extend(result['value'] for result in view.result)
        return results

  def get_paginated_view_results(self,
                                 db_name: str,
                                 design_document_name: str,
                                 view_name: str,
                                 limit: int = 10,
                                 start_key: int | None = None,
                                 start_key_doc_id: str | None = None) -> list[dict[str, Any]]:
    """
    Retrieves paginated view results.

    Args:
        db_name (str): The name of the database.
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.
        limit (int, optional): The maximum number of results to retrieve. Defaults to 10.
        start_key (int | None, optional): The starting key for pagination. Defaults to None.
        start_key_doc_id (str | None, optional): The starting key document ID for pagination. Defaults to None.

    Raises:
        DatabaseError: If design_document_name or view_name is None.

    Returns:
        list[dict[str, Any]]: The paginated view results.
    """
    self.logger.info(
      "Retrieving paginated view results, View: %s from design document: %s, limit: %s, start_key_doc_id: %s, start_key: %s in database: %s",
      view_name,
      design_document_name,
      limit,
      start_key_doc_id,
      start_key,
      db_name)
    params: dict[str, Any] = {
      "limit": limit,
      "descending": True,
    }
    if start_key_doc_id:
      params["startkey_docid"] = start_key_doc_id
    if start_key:
      params["startkey"] = start_key
    if design_document_name is None or view_name is None:
      raise log_and_create_error(self.logger, DatabaseError, "Design document name and view name cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        result = Result(
          design_doc.get_view(view_name),
          **params
        )
        return result[:limit]

  def create_query_index(self,
                         db_name: str,
                         index_name: str,
                         index_type: str = "json",
                         fields: list[dict[str, str]] | list[str] | None = None) -> None:
    """
    Creates a query index in database with the specified name, type, and fields.

    Args:
        db_name (str): The name of the database.
        index_name (str): The name of the index to be created.
        index_type (str, optional): The type of the index. Defaults to "json".
        fields (list[dict[str, str]] | list[str] | None, optional): The fields to be included in the index. Defaults to None.

    Raises:
        DatabaseError: If index_name or fields are None.

    Examples:
        create_query_index("index1", "json", [{"name": "field1", "type": "asc"}])
    """

    self.logger.info("Creating query index, name: %s, index_type: %s, fields: %s in database: %s",
                     index_name,
                     index_type,
                     fields,
                     db_name)
    if index_name is None or fields is None:
      raise log_and_create_error(self.logger, DatabaseError,
                                 "Index name and fields cannot be empty!")
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      if index_name in [index.name for index in pasta_db.get_query_indexes()]:
        self.logger.warning("Index already exists: %s", index_name)
        return
      pasta_db.create_query_index(index_name=index_name,
                                  index_type=index_type,
                                  fields=fields)

  def get_paginated_upload_model_query_results(self,
                                               db_name: str,
                                               filter_term: str | None = None,
                                               filter_fields: list[str] | None = None,
                                               bookmark: str | None = None,
                                               limit: int = 10) -> dict[str, Any]:
    """
    Gets paginated upload model query results based on the provided filter criteria.

    Args:
        db_name (str): The name of the database.
        filter_term (str | None, optional): The filter term to search for. Defaults to None.
        filter_fields (list[str] | None, optional): The list of fields to apply the filter on. Defaults to None.
        bookmark (str | None, optional): The bookmark for pagination. Defaults to None.
        limit (int, optional): The maximum number of results to retrieve. Defaults to 10.

    Returns:
        dict[str, Any]: A dictionary containing the paginated query results. Contains the following keys: bookmark, docs. docs is a list of dictionaries representing the retrieved documents.

    Raises:
        DatabaseError: If filter_term or filter_fields is None.

    Examples:
        get_paginated_upload_model_query_results(filter_term="test", filter_fields=["project_name", "dataverse_url"])

    Raises:
        No specific exceptions are raised by this function.
    """

    self.logger.info(
      "Getting paginated upload model query results: filter_term: %s, filter_fields: %s, bookmark: %s, limit: %s from database: %s",
      filter_term,
      ",".join(filter_fields) if filter_fields else None,
      bookmark,
      limit,
      db_name)
    selector: dict[str, Any] = {"data_type": "dataverse_upload"}
    if filter_term and filter_fields:
      filter_criteria = {"$or": [{field: {"$regex": f"(?i).*{filter_term}.*"}} for field in filter_fields]}
      selector = {
        "$and": [selector, filter_criteria]
      }
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      pasta_db = client[db_name]
      query = Query(pasta_db,
                    selector=selector,
                    sort=[{
                      "finished_date_time": "desc"
                    }],
                    limit=limit)
      return query(bookmark=bookmark) if bookmark else query()

  def get_view_results_by_id(self,
                             db_name: str,
                             design_document_name: str,
                             view_name: str,
                             document_id: str,
                             map_func: str | None = None,
                             reduce_func: str | None = None) -> Document | None:
    """
    Gets the result of a view for a specific document ID from the specified design document in the database.

    Args:
        db_name (str): The name of the database.
        design_document_name (str): The name of the design document.
        view_name (str): The name of the view.
        document_id (str): The ID of the document.
        map_func (str | None, optional): The map function for the view. Defaults to None.
        reduce_func (str | None, optional): The reduce function for the view. Defaults to None.

    Raises:
        ValueError: If design_document_name, view_name, or document_id is None.

    Returns:
        Document | None: The result of the view for the specified document ID, or None if not found.

    Examples:
        # Get the result of a view for a specific document ID
        result = get_view_results_by_id("my_design_doc", "my_view", "my_document_id")

        # Get the result of a view for a specific document ID with map and reduce functions
        result = get_view_results_by_id("my_design_doc", "my_view", "my_document_id", "function(doc) { emit(doc._id, 1); }", "_count")
    """
    self.logger.info(
      "Getting view result: %s for id: %s from design document: %s, map_func: %s, reduce_func: %s from database: %s",
      view_name,
      document_id,
      design_document_name,
      map_func,
      reduce_func,
      db_name)
    with couchdb(self.username,
                 self.password,
                 url=self.url,
                 connect=True) as client:
      if design_document_name is None or view_name is None or document_id is None:
        raise log_and_create_error(self.logger, DatabaseError,
                                   "Design document name, view name and document id cannot be empty!")
      pasta_db = client[db_name]
      with DesignDocument(pasta_db, design_document_name) as design_doc:
        view = View(design_doc, view_name, map_func, reduce_func)
        return view.result[document_id][0]['value']



"""Change the format of dictionaries"""
import re, uuid, json
from typing import Any
from datetime import datetime

def dataHierarchy2Labels(dataHierarchy:dict[str,Any], tableFormat:dict[str,Any]) -> dict[str,Any]:
  """
  Extract labels and create lists of docType,docLabel pair
  - docLabel is the plural human-readable form of the docType
  - docType is the single-case noun

  Not sure if separation into datalabels and hierarchy labels is still needed. Join

  Args:
     dataHierarchy (dict): data hierarchy
     tableFormat (dict): tableFormat branch from .pastaELN.json

  Returns:
     dict: dictionary
  """
  dataDict = {}
  hierarchyDict = {}
  for key in dataHierarchy:
    if key in ['_id', '_rev']:
      continue
    label = None
    if key in tableFormat and '-label-' in tableFormat[key]:  #use label from tableFormat
      label = tableFormat[key]['-label-']
    elif key[0]=='x':                                         #use default structural elements
      label = ['Projects','Tasks','Subtasks','Subsubtasks'][int(key[1])]
    else:                                                     #default system  sample->Samples
      label = key[0].upper()+key[1:]+'s'
    if key[0]=='x':
      hierarchyDict[key] = label
    else:
      dataDict[key] = label
  dataDict |= hierarchyDict
  return dataDict


def dataHierarchy_pre_to_V4(dataHierarchy: dict[str, Any]) -> None:
  """ Translate old dataHierarchy version to version 4

  Based on work of Jithu Murugan

  Args:
      dataHierarchy (dict[str, Any]): dataHierarchy

  """
  icons = {'measurement':'fa5s.thermometer-half', 'sample':'fa5s.vial', 'procedure':'fa5s.list-ol',
           'instrument': 'ri.scales-2-line'}
  shortcuts = {'measurement':'m', 'sample':'s', 'procedure':'p', 'instrument':'i', 'x0':'space'}
  dataHierarchy["-version"] = 4
  typeStructures = {}
  for key in dataHierarchy:
    if key[0] not in {'-', '_'}:
      typeStructures[key] = dataHierarchy[key]
  for docType, typeStructure in typeStructures.items():
    # Adjustments previous versions <= v3.0
    typeStructure.setdefault("attachments", [])
    typeStructure.setdefault("icon", icons[docType] if docType in icons else "")
    typeStructure.setdefault("shortcut", shortcuts[docType] if docType in shortcuts else "")
    properties = typeStructure.get("prop")
    if properties is None:
      properties = {"default": []}
      typeStructure["prop"] = properties
    if not isinstance(properties, dict):
      typeStructure["prop"] = {"default": properties}
    # Adjustments for v4.0
    if "prop" in typeStructure:
      # replace "meta" with "prop" only if it does not exist
      typeStructure.setdefault("meta", typeStructure["prop"])
      del typeStructure["prop"]
    if "label" in typeStructure:
      typeStructure.setdefault("title", typeStructure["label"])
      del typeStructure["label"]
    typeStructure.setdefault("title", "")
  return


def fillDocBeforeCreate(data:dict[str,Any], docType:list[str]) -> dict[str,Any]:
  """ Fill the data before submission to database with common data
  - type, project, children
  - separate comment into tags, fields
  - create id if needed

  used in backend.py and Store.js

  Args:
    data (dict): document to save
    docType (list): type of document, e.g. project=['x0']

  Returns:
    str: document
  """
  protectedKeys = ['comment','-tags','image']
  # Handle the important entries: -type, _id, -date, -branch, -gui
  if '-type' not in data:
    data['-type'] = docType
  if data['-type']==['']:
    data['-type']=['-']
  if isinstance(data['-type'], str):
    data['-type'] = data['-type'].split('/')
  if '_id' not in data:  # if new (if not update): create new id
    data['_id'] = data['-type'][0][0]+'-'+uuid.uuid4().hex
  data['-date']   = datetime.now().isoformat()
  if '-branch' not in data:
    print('Empty branch in data')
    data['-branch'] = [{'stack':[], 'path':None, 'child':-1, 'show':[]}]
  if 'show' not in data['-branch'] and isinstance(data['-branch'], dict):
    data['-branch']['show'] = [True]*(len(data['-branch']['stack'])+1)
  if '-gui' not in data:
    data['-gui'] = [True, True]
  # separate comment into tags and fields
  # these tags are lost: '#d': too short; '#3tag': starts with number
  if 'comment' not in data:
    data['comment'] =''
  if '-tags' not in data:
    data['-tags'] = []
  #always do regex expressions twice: if #lala at beginning or in middle of comment
  curated = re.findall(r'(?:^|\s)#_curated(?:\s|$)', data['comment']) # #_curated
  rating  = re.findall(r'(?:^|\s)#_\d(?:\s|$)',      data['comment']) # #_number
  if rating is None:
    rating=[]
  if len(rating)>1:  #prevent multiple new ratings
    rating=rating[:1]
  if len(rating)==1: #remove ratings that exist already
    data['-tags'] = [i for i in data['-tags'] if not re.compile(r'^_\d$').match(i)]
  otherTags = re.findall(r'(?:^|\s)#[a-zA-Z]\w+(?=\s|$)', data['comment'])
  if otherTags is None:
    otherTags=[]
  data['-tags'] = rating + data['-tags'] + otherTags + curated
  data['comment'] = re.sub(r'(?:^|\s)#\w+(?=\s|$)', '', data['comment']).strip()
  fields = re.findall(r':[\S]+:[\S]+:', data['comment'])
  if fields is not None:
    for item in fields:
      aList = item.split(':')
      if aList[1] in data: #do not add if item already exists
        continue
      data[aList[1]] = aList[2]
  data['comment'] = re.sub(r':[\S]+:[\S]+:','',data['comment'])  #remove :field:data: information
  if isinstance(data['-tags'], str):
    data['-tags'] = data['-tags'].split(' ')
  data['-tags'] = [i.strip()[1:] if i.strip()[0]=='#' else i.strip() for i in data['-tags']]
  data['-tags'] = list(set(data['-tags']))  #ensure only one is inside
  #other cleaning
  if ('links' in data and isinstance(data['links'], list)) and \
     (len(data['links'])==0 or (len(data['links'])==1 and data['links'][0]=='')):
    del data['links']
  #individual verification of documents
  if data['-type'][0]=='sample':
    if 'qrCode' not in data:
      data['qrCode']=[]
    if isinstance(data['qrCode'], str):
      data['qrCode'] = data['qrCode'].split(' ')
  if data['-type'][0] == 'measurement':
    if 'image' not in data:
      data['image'] = ''
    if 'shasum' not in data:
      data['shasum']=''
  # cleaning at end of all changes
  toDelete = []
  for key in data:
    if isinstance(data[key], str):
      if data[key]=='' and key not in protectedKeys:
        toDelete.append(key)
      else:
        data[key] = data[key].strip()
  for key in toDelete:
    del data[key]
  return data


def dict2ul(aDict:dict[str,Any]) -> str:
  """
  convert a dictionary into a corresponding <ul>-html list

  Args:
    aDict (dict): dictionary to be translated
  """
  text = '<ul>'
  for key, value in aDict.items():
    if isinstance(value,str) and value.startswith('{') and value.endswith('}'):
      try:
        value = json.loads(value.replace("'",'"'))
      except:
        pass
    if isinstance(value, dict):
      valueString = dict2ul(value)
    elif isinstance(value, list):
      valueString = ',&nbsp;&nbsp;&nbsp;'.join([str(i) for i in value])
    else:
      valueString = str(value)
    text += f'<li>{key}: {valueString}</li>\n'
  return text+'</ul>'


def diffDicts(dict1:dict[str,Any], dict2:dict[str,Any]) -> str:
  """ Check if two dictionaries differ. Just compare the lowest level keys/values and output string

  Args:
    dict1 (dict): dictionary 1
    dict2 (dict): dictionary 2

  Returns:
    str: output with \\n
  """
  ignoreKeys = ['-client', '_rev', '-gui', '_attachments']
  outString = ''
  dict2Copy = dict(dict2)
  for key,value in dict1.items():
    if key in ignoreKeys:
      continue
    if key not in dict2Copy:
      outString += f'key not in dictionary 2: {key}' + '\n'
      continue
    if value != dict2Copy[key]:
      if key=='-branch':
        if len(value) != len(dict2Copy[key]):
          outString += 'branches have different lengths\n   '+str(value)+'\n   '+str(dict2Copy[key])+'\n'
        else:
          for idx,_ in enumerate(value):
            branch1 = value[idx]
            branch2 = dict2Copy['-branch'][idx]
            if 'show' in branch1:
              del branch1['show']
            if 'show' in branch2:
              del branch2['show']
            if branch1!=branch2:
              outString += 'branches differ\n   '+str(value)+'\n   '+str(dict2Copy[key])+'\n'
      else:
        outString += (f'values differ for key: {key}\n   {str(value)}\n   {str(dict2Copy[key])}\n')
    del dict2Copy[key]
  for key in dict2Copy:
    if key in ignoreKeys:
      continue
    outString += f'key not in dictionary 1: {key}\n'
  return outString
